/*************************************************************************/
/********************** (C) 1991 by Miroslav Kolar ***********************/
/*									 */
/*	          Currently available potentials for 1D.c		 */
/*									 */
/*************************************************************************/
#include <stdio.h>
#include <math.h>
#include <string.h>

#ifndef ultrix
/*      fseek() values  */
#define SEEK_SET 0
#define SEEK_CUR 1
#define SEEK_END 2
#define CLOCKS_PER_SEC 1000000
#endif

#define PR	   (void)fprintf(Listing,
#define PE	   (void)fprintf(stderr,
#define PW	   (void)fprintf(where,
#define SC	   (void)scanf(
#define SI	   (void)fscanf(Input,

extern void exit();

extern FILE *Listing, *Input;
extern double C1, C2, C3, Cx, CE, beta, m;
extern double (*V)();   /* pointer to the selected potential function */
extern void (*inV)();   /* pntr to function for input of the
			   parameters of the selected potential       */
extern long (*outV)();  /*     .......      for output    ......      */

#ifdef QUANTUM
extern double (*dV)();	   /* pointer to dV(x)/dx			  */
extern double (*epsVir)(); /* pointer to the function returning
			V(x) + 0.5*(x-Shift)* dV(x)/dx  for the selected V(x),
			where Shift is nonzero for a potential defined in
			finite interval only, and is selected so that
			the virial estimator gives correct averages	  */
/* Used for approximate methods only: */
	double  (*V2_4)(),   /* pointer to  V''(x) / 4			  */
		(*V23_2)();  /* pointer to the function returning
		   (V''(x) + (x-Shift)* V'''(x)/4) / 2 for the selected V(x) */
	int HOpotFlag = 1;   /* equal to the value of potFlag corresponding
				to the Harmonic Oscillator */
#endif

extern float xMax, xMin;
extern int N_fict;

extern float  xRet1, xRet2;  /* (approximate) classical return points of the
			  potential for energy = Vminimum + 8 / max(beta,1),
			  or some other suitable limits for the generation
			  of the initial configuration and of the
			  histogram range (in atomic units).
			  Must be set in  (*inV)() .			   */

extern void err_fseek(), err_fread();  /* to report errors */

/* #define DEBUG */

/*#########################################################################*/
/* When adding new potentials, don't forget to update this section:        */
/* ---------------------------------------------------------------         */

#define NUMPOT     3     /*  = the number of defined potentials  */

/*     0	   1	      2		*/
void V93init(), VHOinit(), VQOinit();
static void   (*Vinit_all[NUMPOT])() = { V93init, VHOinit, VQOinit };
   /* pointers to functions performing potential dependent initializations */
int  numPot = NUMPOT;   /* total number of currectly defined potentials    */
char *potBlurb, 	/* potential description for printParams()         */
     *potLabel[NUMPOT] = { "9-3", "HO", "QO" }, /* command line, help(),.. */
     *potHelp[NUMPOT] = { "  (12-6 on surface)",
			  "  (Harmonic oscillator)",
			  "  (Quartic oscillator)" };  /* for help()  */
char Label[24];  /* Label[] must be able to hold any element of *potLabel[] */
/*#########################################################################*/

size_t sd = sizeof(double);


void check_label(potFlag)  int potFlag;
/*===================================*/
{
 SI "%s", Label);
#ifdef DEBUG
 PE"check_label: potFlag = %d, Label: %s, potLabel[%d]: %s\n",
	 potFlag, Label, potFlag, potLabel[potFlag]);
#endif
 if( strcmp(Label,potLabel[potFlag]) )
  { PR
    "\nInput file corrupted:  potFlag  is inconsistent with  *potLabel!!!\n");
    exit(1);
  }
}



/*				POTENTIAL # 0				 */
/*  9-3 potential ------------------------------------------- beginning  */

double z0 = 3.0, Vmin = 4.8, C1p, C2p, qCutOff, qCutOffAt;

double V93(q)  double q;
/*====================*/
{
  if(q <= 0. || q > qCutOffAt) return(1e25);
  else { double q3 = q * q * q;

	 return( (C1p / (q3 * q3) - C2p) / q3 );
       }
} 


#ifdef QUANTUM

double dV93(q)  double q;
/*=====================*/
{
  if(q <= 0. || q > qCutOffAt) return(1e25);
  else { double q3 = q * q * q;

	 return ( 3. * (C2p - 3. * C1p / (q3 * q3)) / (q * q3) );
       }
 }

double epsVir93(q)  double q;
/*=========================*/
/*   Here Shift = qCutOff  */
{ 
  if(q <= 0. || q > qCutOffAt) return(1e25);
  else { double q3 = q * q * q, rat = 3. * qCutOffAt / q;

	 return ( 0.5 * (C2p * (1. - rat) -
			 C1p * (7. - 3. * rat) / (q3 * q3)) / q3 );
		/* corrected eps_vir */

/*  else { double q3 = q * q * q;

	 return ( 0.5 * (C2p - 7. * C1p / (q3 * q3)) / q3 );
  /* this is the original version: for Shift=0 */
       }
}

/*
double V93_2_4(q)  double q;
{
  if(q <= 0. || q > qCutOffAt) return(1e25);
  else { double q5 = q * q * q * q * q,
		A = 7.5 * C1p / (q5 * q) - C2p;

	 return ( A > 0 ? 3. * A / q5 : 0. );
       }
 }
*/

double V93_2_4(q)  double q;
/*========================*/
{
  if(q <= 0. || q > qCutOffAt) return(1e25);
  else { double q5 = q * q * q * q * q;

	 return ( 3. * (7.5 * C1p / (q5 * q) - C2p) / q5 );
       }
 }

double V93_23_2(q)  double q;
/*=========================*/
{
  if(q <= 0. || q > qCutOffAt) return(1e25);
  else {
	 double q5 = q * q * q * q * q;
	 return
	   ( 1.5 * (C2p - 52.5 * C1p / (q5 * q)) / q5 );
       }
 }

#endif


void inV93(off)  long off;
/*======================*/
{
  char *p = "inV93";

  if( fseek(Input, off, SEEK_SET) ) err_fseek(p);
#ifdef ultrix
  if( ! fread((void *)&z0, sd, 1, Input) ||
      ! fread((void *)&Vmin, sd, 1, Input) ||
      ! fread((void *)&m, sd, 1, Input) ||
      ! fread((void *)&qCutOffAt, sd, 1, Input) ||
      ! fread((void *)&C1p, sd, 1, Input) ||
      ! fread((void *)&C2p, sd, 1, Input) ) err_fread(p);
#else
  if( ! fread((char *)&z0, sd, 1, Input) ||
      ! fread((char *)&Vmin, sd, 1, Input) ||
      ! fread((char *)&m, sd, 1, Input) ||
      ! fread((char *)&qCutOffAt, sd, 1, Input) ||
      ! fread((char *)&C1p, sd, 1, Input) ||
      ! fread((char *)&C2p, sd, 1, Input) ) err_fread(p);
#endif
  qCutOff = Cx * qCutOffAt;
  check_label(0);
/* PE "z0 = %g, Vmin = %g, m = %g\n", z0, Vmin, m); */
}

long outV93(where)  FILE *where;
/*============================*/
{
  long off = -1.;

  if(where != Listing)
    { off = ftell(where);
#ifdef ultrix
      (void)fwrite((void *)&z0, sd, 1, where);
      (void)fwrite((void *)&Vmin, sd, 1, where);
      (void)fwrite((void *)&m, sd, 1, where);
      (void)fwrite((void *)&qCutOffAt, sd, 1, where);
      (void)fwrite((void *)&C1p, sd, 1, where);
      (void)fwrite((void *)&C2p, sd, 1, where);
#else
      (void)fwrite((char *)&z0, sd, 1, where);
      (void)fwrite((char *)&Vmin, sd, 1, where);
      (void)fwrite((char *)&m, sd, 1, where);
      (void)fwrite((char *)&qCutOffAt, sd, 1, where);
      (void)fwrite((char *)&C1p, sd, 1, where);
      (void)fwrite((char *)&C2p, sd, 1, where);
#endif
      PW"\n");
    }
  PW potLabel[0]);
  PW "   z0 = %.8g A, Vmin = %.8g meV, m = %.10g el mass\n", z0, Vmin, m);
  if(where == Listing) { if(qCutOff == 0.)
				PW"      zCutOff = %.15g * Cx\n", qCutOffAt);
			   else PW"      zCutOff = %.15g\n", qCutOff);
			 PW"      C1p = %.15g, C2p = %.15g\n", C1p, C2p);
		       }
  return(off);
}

void V93init(initFlag)  int initFlag;
/*=================================*/
{
  char flg[2];
  double zsv, betaAt, aux;
  double mDef = 1837.15152; /* (p + e) */  /* mDef = 3675.85; p + n + e */
  double pow();
  char *descript =
      "\nPotential:\t 9-3:   V(x) = 0.5 * Vmin * [(z0/x)^9 - 3 * (z0/x)^3]\n";
  potBlurb = descript;
  V = V93;
  inV = inV93;
  outV = outV93;
#ifdef QUANTUM
  dV = dV93;
  epsVir = epsVir93;
  V2_4 = V93_2_4;
  V23_2 = V93_23_2;
#endif

 if( ! initFlag )
  {
  PE descript);
  m = mDef;

  PE "\nDefault mass =  %.10g el masses. Change? (y/n) ... ", m);
  SC "%s", flg);
  if (flg[0] == 'y') { PE "\nEnter mass ... ");
		       SC "%lf", &m);
		     }
  PE "\nDefault z0 =  %.8g Angstrom. Change? (y/n) ... ", z0);
  SC "%s", flg);
  if (flg[0] == 'y') { PE "\nEnter z0 ... ");
		       SC "%lf", &z0);
		     }
  PE "\nDefault Vmin =  %.8g meV. Change? (y/n) ... ", Vmin);
  SC "%s", flg);
  if (flg[0] == 'y') { PE "\nEnter Vmin ... ");
		       SC "%lf", &Vmin);
		     }

  PE "\nEnter  z  cut off value (Angstrom) ... ");
  SC "%lf", &qCutOff);
  PE"\n");

  zsv =  z0 * sqrt(Vmin * m);
  C1p = 0.33765785681543063742e-8 * pow(zsv,5.5);
  C2p = 4.8027155728271075574e-5 * pow(zsv,2.5);
  zsv =  z0 / sqrt(Vmin);
  aux = sqrt(m);
  Cx = 4.0987188166381507849 * sqrt(zsv / aux);
  CE = 726.73121533124813990 / (1.6021892 * zsv * aux);
  qCutOffAt = qCutOff / Cx;
  /* All plotting and printing is in  meV  and  Angstrom */
#ifdef QUANTUM
	betaAt = beta * CE;
	C1 = N_fict / (2. * betaAt);
	C2 = C1 / betaAt;
	C1 = C1 / M_PI;
#endif
  xRet2 = 3. * z0 / Cx;
  if(xRet2 > qCutOffAt) xRet2 = qCutOffAt;
  xRet1 = pow(C1p / C2p, 0.16666666666666666);
  if(xRet2 <= 4. * xRet1) xRet2 = 4. * xRet1;
  if(xMax < -1e+29) xMax = qCutOff;
  if(xMin > 1e+29)  xMin = 0.95 * xRet1 * Cx;
  if(xMax > qCutOff) xMax = qCutOff;
  }
 }

/*  9-3 potential ------------------------------------------------- end  */


/*				POTENTIAL # 1				 */
/*  Harmonic oscillator -------------------------------------- beginning
    Everything is done only in atomic units: Cx=CE=1                     */

double VHO(x)  double x;
/*====================*/
{  return(0.5 * x * x);  }


#ifdef QUANTUM

double dVHO(x)  double x;
/*=====================*/
{  return(x);  }

double epsVirHO(x)  double x;
/*=========================*/
/*   Here Shift = 0 (no need for infinite space, no help for symmetric box)  */
{  return(x * x);  }

double VHO_2_4(x)  double x;
/*========================*/
{  return(0.25);  }

double VHO_23_2(x)  double x;
/*=========================*/
{  return(0.5);  }

#endif


void inVHO(off)  long off;
/*======================*/
{
  char *p = "inVHO";
  if( fseek(Input, off, SEEK_SET) ) err_fseek(p);
  check_label(1);
  /* No further parameters needed */
}

long outVHO(where)  FILE *where;
/*============================*/
{
  long off = -1.;

  if(where != Listing)  off = ftell(where);
  PW potLabel[1]); PW"\n");
  return(off);
}

void VHOinit(initFlag)  int initFlag;
/*=================================*/
{
 char *descript = "\nPotential:\t Harmonic oscillator:  V(x) = (1/2) x^2\n";
 potBlurb = descript;
 V = VHO;
 inV = inVHO;
 outV = outVHO;
#ifdef QUANTUM
 dV = dVHO;
 epsVir = epsVirHO;
 V2_4 = VHO_2_4;
 V23_2 = VHO_23_2;
#endif

 if( ! initFlag )
	{ m = 1.;
	  xRet2 = 8. / (beta < 1. ? beta : 1.);
	  xRet1 = -xRet2;
	  if(xMax < -1e+29  ||  xMin > 1e+29)
		{ xMax = xRet2;
		  xMin = -xMax;
		}
#ifdef QUANTUM
	  C1 = N_fict / (2. * beta);
	  C2 = C1 / beta;
	  C1 = C1 / M_PI;
#endif
	  Cx = CE = 1.;
	}
}

/*  Harmonic oscillator ------------------------------------------- end  */
 

/*				POTENTIAL # 2				 */
/*  Quartic oscillator -------------------------------------- beginning
    Everything is done only in atomic units: Cx=CE=1			 */

double VQO(x)  double x;
/*====================*/
{ double xx = x * x; 
  return(0.5 * xx * xx);
}


#ifdef QUANTUM

double dVQO(x)  double x;
/*=====================*/
{ return(2. * x * x * x);
}

double epsVirQO(x)  double x;
/*   Here Shift = 0 (no need for infinite space, no help for symmetric box)  */
/*=========================*/
{ double xx = x * x; 
  return(1.5 * xx * xx);
}

double VQO_2_4(x)  double x;
/*======================*/
{  return(1.5 * x * x);  }

double VQO_23_2(x)  double x;
/*=======================*/
{  return(4.5 * x * x);  }

#endif


void inVQO(off)  long off;
/*======================*/
{
  char *p = "inVQO";
  if( fseek(Input, off, SEEK_SET) ) err_fseek(p);
  check_label(2);
  /* No further parameters needed */
}

long outVQO(where)  FILE *where;
/*============================*/
{
  long off = -1.;

  if(where != Listing)  off = ftell(where);
  PW potLabel[2]); PW"\n");
  return(off);
}

void VQOinit(initFlag)  int initFlag;
/*=================================*/
{
 char *descript = "\nPotential:\t Quartic oscillator:  V(x) = (1/2) x^4\n";
 potBlurb = descript;
 V = VQO;
 inV = inVQO;
 outV = outVQO;
#ifdef QUANTUM
 dV = dVQO;
 epsVir = epsVirQO;
 V2_4 = VQO_2_4;
 V23_2 = VQO_23_2;
#endif

 if( ! initFlag )
	{ m = 1.;
	  /* xRet2 = 1.414 * pow(beta,0.25); */
	  xRet2 = 8. / (beta < 1. ? beta : 1.);
	  xRet1 = -xRet2;
	  if(xMax < -1e+29  ||  xMin > 1e+29)
		{ xMax = 1.2 * xRet2;
		  xMin = -xMax;
		}
#ifdef QUANTUM
	  C1 = N_fict / (2 * beta);
	  C2 = C1 / beta;
	  C1 = C1 / M_PI;
#endif
	  Cx = CE = 1.;
	}
}

/*  Quartic oscillator ------------------------------------------- end  */
 


void choosePotential(potFlag,initFlag)  int potFlag,initFlag;
/*=========================================================*/
/* initFlag = 0   :  a new calculation is started; all potential
		     parameters must be calculated.
   initFlag != 0  :  it is assumed that the potential parameters were
		     already read into the calling program from a file.
		     Thus they needn't be recalculated.
*/
{
#ifdef DEBUG
  PE"\n\nchoosePotential:  potFlag: %d\n\n", potFlag);
#endif
  if(potFlag < 0 || potFlag >= NUMPOT)
	{ PR"\n\nchoosePotential: Illegal value of potFlag: %d\n\n", potFlag);
	  exit(1);
	}
  (*Vinit_all[potFlag])(initFlag);
#ifdef QUANTUM
  C3 = 2.* C2;
#endif
}
