/************************************************************************/
/*    This program calculates the Davis-Rabinowitz test integral	*/
/*    [see D & R: "Methods of Numerical Integration", 1984, Acad.Press;	*/
/*     and   B. L. Fox, ACM Trans.Math.Soft. 12, 362 (1986)].		*/
/*	    At present for Halton and Ranmar generators.		*/
/*									*/
/*			   M. Kolar 1992				*/
/************************************************************************/

#include <errno.h>
/*#include <time.h>*/
#include <sys/times.h>
#include "1D.h"

/*
extern void exit();
#ifdef ultrix
extern void *malloc();
#else
extern char *malloc();
#endif
*/

FILE *Listing = stdout, *Output = stdout, *Density = stdout, *Input;

void err_fread(flg)  int *flg;
/*==========================*/
{  *flg = 0;
   if(feof(Input))  PR"\n End of Input file\n");
        else        PR"\7\n fread error\n");
}

/*-----------------------------------------------------------------------*/
/* Currently available random number generators are defined in RanGen.c. */
/* They require the following declarations:				 */

int rg1Flag, rg2Flag;  /* indicate the two selected generators for
			 particle movement and configuration acceptance	*/
char *rg1 = "\0", *rg2 = SINGLE_STREAM_STRING;
	 /* pointers to verbal descriptions of the two RNG's */

double randNum1[1], randNum2[1];
	/* buffers to store N_fict random numbers generated by the 
	   two generators; these numbers are just enough for one MC cycle */

extern void out_RnGn_state();  /* formated output of the state vectors  */

extern int init_rand_gen();  /* selects and initializes the two generators */

extern double (*RnGn1)(), (*RnGn2)();
	  /* pointers to the two selected RNG */

extern void Eratosthenes();

/*	End of definitions for the file RanGen.c			 */
/*-----------------------------------------------------------------------*/

/* RANMAR from file ranmar1.c is used here: */
extern int rmarin1();
extern double ranmar1();

/*-----------------------------------------------------------------------*/
/* The basic algorithms for the Halton numbers as presented
   in the paper "Fast, portable, and reliable algorithm for ...."
   are in the file RanGen1.c.
   It also contains the generator of the full Halton vector using the 
   first nBase primes, and an unsuccesful "experimental generator"
   based on binary Halton.
*/

extern double init_HaltonBS();
extern double HaltonBS();
extern void init_Halton();
extern double Halton();
extern void init_Halton2();
extern double Halton2();
extern void init_vHalton();
extern void rand_init_vHalton();
extern double vHalton();
extern void init_ExprG();
extern double ExprG();
extern int N0[];
/*	End of definitions for the file RanGen1.c			 */
/*-----------------------------------------------------------------------*/

#define dim 40	     /*  dimension of the integral  */
/* #define Gm   8       /*  number of iterations of N  */
/* #define A1m  200     /*  first averaging */
/* #define A2m  10      /*  second averaging */


main(agc, agv) int agc; char *agv[];
/*================================*/
{
  FILE *fAvrg, *fStd, *fSkew;
  int i, j, k, l, m, flg, A1 = 100, A2 = 5, A3, a2, a3, G, N = 1000, M;
				/*  N = max. number of grid points  */
  int seed = 32737;
  long ij, kl;
/*  double Int[A1m][Gm], av[A3m][Gm], Av[Gm], Var[Gm], std[A3m][Gm]; */
  double **Int, **av, *Av, *Var, **std, *sum, *InInt, *skew;
  double S, P, (*pR)(), aux;
  extern double pow(), sqrt(), drand48();
  extern long mrand48();
  extern void srand48();
  static unsigned Seed[] = { 1345654, 693125, 134, 1, 71945, 3675, 37 };
				/* Historical reasons */
  clock_t times();
  double usLast, sysLast, uTo, sTo;
  struct tms cpuT;
  char *IntFile = "Integrals";
  int eflg = 1, *p = &eflg;

if(agc == 1 ||
   (agv[1][0] != 'h' && agv[1][0] != 'r' && agv[1][0] != 'R'
		     && agv[1][0] != 's' && agv[1][0] != 'p')
  )
 { PR"    Calculation of a Davis-Rabinowitz test integral\n");
   PR"    see D & R: \"Methods of Numerical Integration\", 1984, Acad.Press;\n");
   PR"        B. L. Fox, ACM Trans.Math.Soft. 12, 362 (1986)\n");
   PR"\nUsage: %s h ['seed' ['A1' ['A2' ['Nmax']]]]    (for Halton)\n",	agv[0]);
   PR"       %s r ['seed' ['A1' ['A2' ['Nmax']]]]    (for Ranmar)\n", agv[0]);
   PR"               A1 is the total # of calculations\n");
   PR"               A2 is the interval between averaging\n");
   PR"               Nmax is the max number of grid points\n");
   PR"  or\n");
   PR"       %s R ['A1' ['A2' ['file']]]\n", agv[0]);
   PR"               ...  Repeat calculation using Integrals from 'file'\n");
   PR"  or\n");
   PR"       %s s ['file' ['col' ['lower' ['upper']]]]    (statistics)\n",
		agv[0]);
   PR"               file   - file containing Integrals\n");
   PR"               col    - column # (corr's to N = 500, 1000, 7000, etc)\n");
   PR"               lower, upper - find # of integrals in the given column\n");
   PR"                        having values between 'lower' and 'upper'\n");
   PR"  or\n");
   PR"       %s p ['file']          (print Integrals)\n", agv[0]);
   exit(0);
 }

if(agv[1][0] == 's' || agv[1][0] == 'p')
 { int in = 0;
   double lower = 0.7873, upper = 1.2127;

   if(agc > 2) IntFile = agv[2];
/*   if( (Input = fopen(IntFile,"rb")) == NULL) */
   if( (Input = fopen(IntFile,"r")) == NULL)
		{ PR"\nCan't open %s\n\n", agv[2]);  exit(-1); }
   RI(&G, 1)

   if(agv[1][0] == 's')
      { if(agc > 3) { j = atoi(agv[3]);
		      if(agc > 4) { lower = atof(agv[4]);
				    if(agc > 5) upper = atof(agv[5]);
				  }
		    }
		    else  j = G;

	if(j > G)  { PR"\n'col' must not be larger than %d\n\n", G);
		     exit(-1);
		   }
	if(lower >= upper) { PR"\n'lower' must be smaller than 'upper'\n\n");
			     exit(-1);
			   }
        j--;
      }
   InInt =  (double *)malloc(G * sizeof(double));

   RI(&M, 1)
   /* If end of file is encountered, eflg (=*p) is set to 0 by RD or RI */
   while(eflg) { RD(InInt, G)
		 if(agv[1][0] == 's')
			{ if(InInt[j] >= lower && InInt[j] <= upper) in++; }
			else { PR"\n%d   ", M);
			       for(j = 0; j < G; j++) PR"%#7.4f ", InInt[j]);
			     }
		 RI(&M, 1)
	       }

   if(agv[1][0] == 'p') PR"\n\n");
   else 
   {
    PR"\n File: %s, column = %d, total # of integrals = %d\n\n",
			agv[2], j + 1, M);
    PR" Between %g and %g there is %d integrals\n", lower, upper, in);
    PR"            i.e. %g %%, outside the remaining %g %%\n\n",
			100. * in / (double)M, 100. * (M - in) / (double)M );
   }
   exit(0);
 } /* end of  "if(agv[1][0] == 's' || agv[1][0] == 'p')"  */


if(agv[1][0] == 'R')
  { if(agc > 2) { A1 = atoi(agv[2]);
		  if(agc > 3) { A2 = atoi(agv[3]);
				if(agc > 4) IntFile = agv[4];
			      }
		}
/*    if( (Input = fopen(IntFile,"rb")) == NULL) */
    if( (Input = fopen(IntFile,"r")) == NULL)
		{ PR"\nCan't open %s\n\n", agv[2]); exit(-1); }
  }
  else
  { if(agc > 2) { seed = atoi(agv[2]);
		  if(agc > 3) { A1 = atoi(agv[3]);
				if(agc > 4) { A2 = atoi(agv[4]);
					      if(agc > 5) N = atoi(agv[5]);
					    }
			      }
		}
    srand48((long)abs(seed));
  }

A3 = (A1 + A2 / 2) / A2;
A1 = A3 * A2;

if(agv[1][0] == 'R')  { RI(&G, 1) }
  else {
	if(N >= 1000000) { N = 1000000; G = 8; }
	  else if(N >= 200000) { N = 200000; G = 7; }
	  else if(N >= 100000) { N = 100000; G = 6; }
	  else if(N >= 40000) { N = 40000; G = 5; }
	  else if(N >= 20000) { N = 20000; G = 4; }
	  else if(N >= 7000) { N = 7000; G = 3; }
	  else if(N >= 1000) { N = 1000; G = 2; }
	  else { N = 500; G = 1; }
/*	Output = fopen(IntFile, "wb"); */
	Output = fopen(IntFile, "w");
	WI(&G,1)
       }

Int = (double **)malloc(A1 * sizeof(double *));
for(m = 0; m < A1; m++) Int[m] = (double *)malloc(G * sizeof(double));
av = (double **)malloc(A3 * sizeof(double *));
for(l = 0; l < A3; l++) av[l] = (double *)malloc(G * sizeof(double));
Av =  (double *)malloc(G * sizeof(double));
Var =  (double *)malloc(G * sizeof(double));
sum =  (double *)malloc(G * sizeof(double));
std = (double **)malloc(A3 * sizeof(double *));
for(l = 0; l < A3; l++) std[l] = (double *)malloc(G * sizeof(double));
skew =  (double *)malloc(G * sizeof(double));

fAvrg = fopen("Averages","w");
fStd = fopen("StDev","w");
fSkew = fopen("Skewness","w");

PR"\nStarted:  "); PR DateAndTime());
PR"\n\n            Calculation of the Davis-Rabinowitz test integral\n");
PR"            =================================================\n");
PR"                  Written by M. Kolar (C) 1992\n\n");

(void)times(&cpuT);
usLast = (double)cpuT.tms_utime / 60.;
sysLast = (double)cpuT.tms_stime / 60.;
PR"Times(): from start: user  %g, system  %g\n\n", usLast, sysLast);

/*   rg1 = agv[3];
   PE"dim = %d      %s\n", dim, rg1);
   init_rand_gen(&rg1, &rg2, 0, 1);
   out_RnGn_state(0,stderr);
*/
/*  int N = atoi(agv[1]), dim = atoi(agv[2]);
  PE"dim = %d\n", dim);
  if(!strcmp(agv[3],"rand")) rand_init_vHalton(dim, atoi(agv[4]));
   else { if(!strcmp(agv[3],"beg")) for(i = 0; i < dim; i++) N0[i] = 1;
	  else for(i = 0; i < dim; i++) N0[i] = atoi(agv[i + 3]);
	  init_vHalton(dim,N0);
	}
*/

if(agv[1][0] == 'R') 
   PR"   Repeating calculation using integrals from file \'%s\'\n\n", IntFile);
   else
   {
     if(flg = (agv[1][0] == 'h')) pR = vHalton; else pR = ranmar1;
     if(flg) PR"\t\t\tvHalton()"); else PR"\t\t\tranmar()");
     PR"    seed = %d\n", seed);
     PR"\t\t\t========\n\n");
   }

PR"dim = %d\n", dim);
PR"               500    1000    7000     2e4     4e4     1e5     2e5\
     1e6\n");
Listing = fAvrg;
PR"           500    1000    7000     2e4     4e4     1e5     2e5     1e6\n");
Listing = fStd;
PR"           500    1000    7000     2e4     4e4     1e5     2e5     1e6\n");
Listing = fSkew;
PR"           500    1000    7000     2e4     4e4     1e5     2e5     1e6\n");
Listing = stdout;

/*
{int Sd;
for(Sd = 0; Sd < 10000; Sd++) { srand48((long)Sd);
*/

for(j = 0; j < G; j++) sum[j] = 0.;
for(l = 0; l < A3; l++) 
    for(j = 0; j < G; j++) av[l][j] = std[l][j] = 0.;
a2 = a3 = 0;

if(agv[1][0] == 'R')   { RI(&M, 1) }
   /* If end of file is encountered, eflg (=*p) is set to 0 by RD or RI */

for(m = 0; m < A1 && eflg; m++)
  { j = -1; S = 0.;

    if(agv[1][0] == 'R')
      {
	/* Read Integrals from Input: */
	RD(Int[m], G)
        for(j = 0; j < G; j++)
	  { av[a3][j] += Int[m][j]; sum[j] += Int[m][j]; }
      }
      else
      {
	M = m + 1;
	if(flg) { /* if(l == 0) rand_init_vHalton(dim, Seed[m]);
		       else */
		  if(seed <= 0 && m == 0)
			 { for(i = 0; i < dim; i++) N0[i] = -seed;
			   init_vHalton(dim, N0);
			 }
		     else  rand_init_vHalton(dim, (unsigned)mrand48());
		}
	else  { ij = drand48() * 31328; 
		kl = drand48() * 30081;
		(void)rmarin1(&ij, &kl);
	      }

#define CYCLE { P = 1; /* P = (2 * i - 1)/ N - 1; Hammersley set: dim - 1 */\
		for(k = 0; k < dim; k++) P *= 2 * (*pR)() - 1.;\
		S += fabs(P);\
		/*  PE"%d %g\n", i, P * pow(2.,(double)dim));  */\
	      }

	for(i = 1; i <= 500; i++) CYCLE
		Int[m][++j] = S * pow(2.,(double)dim) / 500.;
/*
if(fabs(Int[m][j] - 0.1757) <= 0.0001) PE"%d %g\n", Sd, Int[m][j]); 
if(fabs(Int[m][j] - 0.1963) <= 0.0001) PE"%d %g\n", Sd, Int[m][j]); 
*/
		av[a3][j] += Int[m][j]; sum[j] += Int[m][j];
		if(N <= 500) goto BRK;
	for( ; i <= 1000; i++) CYCLE
		Int[m][++j] = S * pow(2.,(double)dim) / 1e3;
		av[a3][j] += Int[m][j]; sum[j] += Int[m][j];
		if(N <= 1000) goto BRK;
	for( ; i <= 7000; i++) CYCLE
		Int[m][++j] = S * pow(2.,(double)dim) / 7e3;
		av[a3][j] += Int[m][j]; sum[j] += Int[m][j];
		if(N <= 7000) goto BRK;
	for( ; i <= 20000; i++) CYCLE
		Int[m][++j] = S * pow(2.,(double)dim) / 2e4;
		av[a3][j] += Int[m][j]; sum[j] += Int[m][j];
		if(N <= 20000) goto BRK;
	for( ; i <= 40000; i++) CYCLE
		Int[m][++j] = S * pow(2.,(double)dim) / 4e4;
		av[a3][j] += Int[m][j]; sum[j] += Int[m][j];
		if(N <= 40000) goto BRK;
	for( ; i <= 100000; i++) CYCLE
		Int[m][++j] = S * pow(2.,(double)dim) / 1e5;
		av[a3][j] += Int[m][j]; sum[j] += Int[m][j];
		if(N <= 100000) goto BRK;
	for( ; i <= 200000; i++) CYCLE
		Int[m][++j] = S * pow(2.,(double)dim) / 2e5;
		av[a3][j] += Int[m][j]; sum[j] += Int[m][j];
		if(N <= 200000) goto BRK;
	for( ; i <= 1000000; i++) CYCLE
		Int[m][++j] = S * pow(2.,(double)dim) / 1e6;
		av[a3][j] += Int[m][j]; sum[j] += Int[m][j];
	BRK:
/*	      { P = 1;
 		for(k = 0; k < dim; k++) P *= (*pR)();
		S += exp(-P);
		if(i == 1000) { Int[m][++j] = S / 1e3;
				av[a3][j] += Int[m][j]; sum[j] += Int[m][j]; }
		if(i == 7000) { Int[m][++j] = S / 7e3;
				av[a3][j] += Int[m][j]; sum[j] += Int[m][j]; }
		if(i == 10000) { Int[m][++j] = S / 1e4;
				 av[a3][j] += Int[m][j]; sum[j] += Int[m][j]; }
		if(i == 12000) { Int[m][++j] = S  / 1.2e4;
				 av[a3][j] += Int[m][j]; sum[j] += Int[m][j]; }
	      }
*/
	if(j >= G) { PR"\n1st dimension of Int[][], ... exceeded!\n\n");
		     exit(-1);
		   }
	/* Write Integrals to Output: */
	WI(&M, 1)
	WD(Int[m], G)
	(void)fflush(Output);
     }

   if(++a2 == A2)
	  {
	    for(j = 0; j < G; j++)
		{ Av[j] = sum[j] / (double)M;
		  Var[j] = 0.;
		  skew[j] = 0.;
		  for(l = 0; l < M; l++) { aux = Int[l][j] - Av[j];
					   Var[j] += aux * aux;
					   skew[j] += aux * aux * aux;
					 }
		}
	    Listing = fAvrg;
	    PR"\n %4d  ", M);
	    for(j = 0; j < G; j++) PR"%#7.3f ", Av[j]);
	    (void)fflush(Listing);
	    Listing = fStd;
	    PR"\n %4d  ", M);
	    for(j = 0; j < G; j++) PR"%#7.3f ", sqrt(Var[j] / (double)M) );
	    (void)fflush(Listing);
	    Listing = fSkew;
	    PR"\n %4d  ", M);
	    for(j = 0; j < G; j++)
	      PR"%#+7.3f ", skew[j] / (M * pow(Var[j] / (double)(M - 1),1.5)) );
	    (void)fflush(Listing);

	    Listing = stdout;
	    PR"\n%4d-%4d: ", M - A2 + 1, M);
	    for(j = 0; j < G; j++)
		{ av[a3][j] /= (double)A2;
		  PR"%#7.3f ", av[a3][j]);
		  Var[j] = 0.;
		  for(l = M - A2; l < M; l++) { aux = Int[l][j] - av[a3][j];
						Var[j] += aux * aux; }
		  Var[j]  /= (double)A2;
		  std[a3][j] = sqrt(Var[j]);
		}
	    PR"\n   +/-     ");
	    for(j = 0; j < G; j++) PR"%#7.3f ", std[a3][j]);
	    (void)fflush(Listing);
	    a2 = 0;
	    a3++;
	  }

    if(agv[1][0] == 'R')  { RI(&M, 1) }
  }  /* end of  "for(m = 0; m < A1 && eflg; m++)"   */

Listing = stdout;
PR"\n\n<<  >>:    ");
for(j = 0; j < G; j++)
  { sum[j] = 0;
    for(l = 0; l < a3; l++) sum[j] += av[l][j];
    Av[j] = sum[j] / (double)a3;
    PR"%#7.3f ", Av[j]);
  }

PR"\n<st dev>:  ");
for(j = 0; j < G; j++)
  { Var[j] = 0;
    sum[j] = 0;
    for(l = 0; l < a3; l++) sum[j] += std[l][j];
    Av[j] = sum[j] / (double)a3;
    PR"%#7.3f ", Av[j]);
  }

PR"\n    +/-    ");
for(j = 0; j < G; j++)
  { for(l = 0; l < a3; l++) { aux = std[l][j] - Av[j];
			      Var[j] += aux * aux; }
    Var[j]  /= (double)a3;
    PR"%#7.3f ", sqrt(Var[j]));
  }

(void)fflush(Listing);
PR"\n\n\nIntegrals: ");
for(m = 0; m < (M > 50 ? 50 : M); m++)
  { PR"\n    %2d   ", m + 1);
    for(j = 0; j < G; j++) PR"%#7.3f ", Int[m][j]);
  }
PR"\n");
/*
}}
*/
(void)fclose(fAvrg);
(void)fclose(fStd);
(void)fclose(fSkew);

   (void)times(&cpuT);
   uTo = (double)cpuT.tms_utime / 60.;
   sTo = (double)cpuT.tms_stime / 60.;
   PR"\n\nTimes(): user  %g, system  %g, total  %g\n",
	uTo - usLast , sTo - sysLast, uTo + sTo - (usLast + sysLast));
   usLast = uTo;
   sysLast = sTo;

PR"\nWhole job: Times(): user  %g, system  %g, total  %g\n",
					uTo, sTo, uTo + sTo);
PR"Finished:  "); PR DateAndTime());
PR"\n\n");

}
