/***************************************************************************/
/********************** (C) 1991 by Miroslav Kolar *************************/
/*                                                                         */
/*	 This is a package to do simple line-printer-type plots 	   */
/*	 ------------------------------------------------------		   */
/* Functions defined: initPlot, xRange, plotDens, plotPoints, plotF,	   */
/*		      plotFun, plotFun1, Histogram			   */
/*									   */
/* initPlot:  initialization						   */
/* -------------------------						   */
/*									   */
/* xRange:  plots x-axis   (scaled by CX, see below)			   */
/* ---------------------						   */
/*									   */
/* plotDens: plots density of 1D points:				   */
/* ------------------------------------					   */
/* On each line, first a quantity (line identification) passed as a first  */
/* parameter to plotDens() is plotted, then in W columns all the points    */
/* that fall into the plotted region ('1' represents one point at the      */
/* given position, '2' represents two overlapping points, etc).            */
/* At the end of the line the number of the points in the plotted region   */
/* is printed								   */
/*									   */
/* plotPoints: plots positions of ordered 1D points:			   */
/* ------------------------------------------------			   */
/* Same as plotDens, only the labels of points are different: they are     */
/* their ordinal numbers: 1,2,...,9,A,B,.....,Z; all others: #		   */
/*									   */
/* plotFun, plotFun1, plotF: X-Y  plots:				   */
/* ------------------------------------					   */
/* The quantity printed on the left is the y-value (scaled by CY).	   */
/* Asterisk '*' is used to represent all points.			   */
/* (The complexity of plotFun is due to the fact that there may be	   */
/*  several values of x for each y)					   */
/*   plotFun:  min and max, Fmin and Fmax, of the plotted function must	   */
/*	       be supplied						   */
/*   plotFun1: Fmin and Fmax are calculated automatically		   */
/*   plotF:    used by plotFun() and Histogram() to plot a single line	   */
/*									   */
/* Histogram: similar to plotFun, plots a histogram			   */
/* ---------								   */
/* (No scaling of the y-values printed on the left)			   */
/*                                                                         */
/***************************************************************************/

#include <stdio.h>
#include <string.h>

#define PR         (void)fprintf(Listing,
#define MAXCOLUMN  120
#define MAXOVRLP   61   /* when increasing MAXOVRLP, new labels must be
			   added to lab[]				 */
#define LINES	   20   /* number of lines for a X-Y plot		 */

extern void exit();
#ifdef ultrix
extern void *malloc(), free();
#else
extern char *malloc();
extern int free();
#endif
  
static FILE *Listing;
static float X1, X2, /* lower and upper limits of the plotting area  */
	     Xstep, /* length of interval (bin) represented by one character */
	     CX, /*  = x_graph / x_internal   | scaling between the internal */
	     CY; /*  = y_graph / y_internal   | values and the plotted ones  */
static int N, /* # of points to plot on one line */
	   W; /* # of bins : Xstep = (X1 - X2) / W;               */
static char *FMT;
	/* format for value to be printed in the beginning of the line  */
static int  FMT_LEN; /* total number of char printed using  FMT  format */

static char lab[MAXOVRLP + 3] = { " 123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ\
abcdefghijklmnopqrstuvwxyz?" };
	/* in plotDens: if a bin contains N values, lab[N] will be printed
		   in the respective position on the line if  N <= MAXOVRLP;
		   if  N > MAXOVRLP, lab[MAXOVRLP + 1] is printed
	   in plotPoints: used as the ordinal number of the plotted point  */
static char lab1[3] = { " *" };  /* for X-Y plots                          */

static short *indb, *ord;  /* Used in plotPoints() */


void initPlot(Cx,Cy,x1,x2,n,w,fmt,fmt_len,file)
     float Cx,Cy,x1,x2; int n,w,fmt_len; char *fmt; FILE *file;
/*===========================================================*/
{ Listing = file;
  X1 = x1;
  X2 = x2;
  CX = Cx;
  CY = Cy;
  indb = (short int *)malloc((unsigned)n * sizeof(short int));
  ord = (short int *)malloc((unsigned)n * sizeof(short int));
  N = n;
  FMT = fmt;
  FMT_LEN = fmt_len;
  W = (w > MAXCOLUMN) ? MAXCOLUMN : w;
  Xstep = (x2 - x1) / W;
}

void xRange()
/*=========*/
/* Indicates the extent of the x-domain being plotted */
{
  int pos, len, len2;
  char aux[20];

  for(pos = 0; pos < FMT_LEN; pos++) PR" ");
  PR"|");
  for(pos = 0; pos < W; pos++) PR" ");
  PR"|\n");
  for(pos = 0; pos < FMT_LEN; pos++) PR" ");
  PR"|%-g <%n", CX * X1, &pos);
  (void)sprintf(aux,"> %g%n|\n", CX * X2, &len);
  len = W - pos - len + 1;
  len2 = len / 2 + 2;
  while(len-- > len2) PR"-");
  PR" x "); len -= 2;
  while(len-- > 0) PR"-");
  PR aux);
}

void plotDens(num,x) long num; double *x;
/*=====================================*/
/* Assumes that plotInit() was called before  */
{
  register short i;
  register short ind;
  short sum = 0;
  short  bin[MAXCOLUMN];   /*  used to generate one output line  */

  lab[0] = ' ';

  for(i = 0; i < W; i++) bin[i] = 0;

  for(i = 0; i < N; i++) { ind = (*x++ - X1) / Xstep;
			   if(ind >= 0 && ind < W) (bin[ind])++;
			 }
  for(i = 0; i < W; i++) { sum += bin[i];
			   if(bin[i] > MAXOVRLP) bin[i] = MAXOVRLP + 1;
			 }
  PR FMT, num); PR"|");
  for(i = 0; i < W; i++) PR"%c", lab[bin[i]]);
  PR "| %d\n",sum);
   (void)fflush(Listing);
}

void plotPoints(num,x) long num; double *x;
/*=======================================*/
/* Assumes that plotInit() was called before  */
{
  register short i;
  register short ind;
  register short l;
  register short l1;
  register short k;
  register short s;
  /*  indb  -  stores the index of the bin into which a given point falls */
  /*  ord   -  stores the ordinal number of that point			  */
  /*  ordering: indb[0] <= indb[1] <= ... indb[s];			  */
  /*		s+1 <= N is the number of points contained in the plotted */
  /*		range							  */

  lab[0] = '0';

  s = -1;
  for(i = 0; i < N; i++) { ind = (*x++ - X1) / Xstep;
			   if(ind >= 0 && ind < W)
				 { for(k = 0; k <= s; k++) 
					{ if(ind < indb[k]) break; }
				   l = s;
				   while(l >= k) { l1 = l + 1;
						   indb[l1] = indb[l];
						   ord[l1] = ord[l];
						   l--;
						 }
				   indb[k] = ind;
				   ord[k] = i;
				   s++;
				 }
			 }
  for(i = 0; i <= s; i++) if(ord[i] > MAXOVRLP) ord[i] = MAXOVRLP + 1;
  PR FMT, num); PR"|");
  if(s >= 0) {
		k = indb[0];
		for(i = 0; i < k; i++) PR" ");
		PR"%c", lab[ord[0]]);
		for(l = 1; l <= s; l++)
			{ ind = indb[l];
			  if(ind == k) PR"\b");
				else  { while(++k < ind) PR" ");
					k = ind;
				      }
			  PR"%c", lab[ord[l]]);
			}
	     }
	     else k = -1;
  while(++k < W) PR" ");
  PR "| %d\n", s + 1);
  (void)fflush(Listing);
}

  
/*    -------------------------   X-Y plots:  ------------------------    */

void plotF(y,n,x,fmt) float y, *x; int n; char *fmt;
/*================================================*/
/* Used by plotFun() and Histogram(), see below  */
/* n  is the number of x-values corresponding to  y, and pointed to by x  */
/* Here, y is assumed to be already the value that is actually printed    */
{
  char aux[20];
  register short i;
  register short ind;
  short  bin[MAXCOLUMN];   /*  used to generate one output line  */

  for(i = 0; i < W; i++) bin[i] = 0;

  for(i = 0; i < n; i++) { ind = (*x++ - X1) / Xstep;
			   if(ind >= 0 && ind < W) bin[ind] = 1;
			 }
  (void)sprintf(aux, fmt, y); aux[FMT_LEN] = '\0';
  PR aux); PR"|");
  for(i = 0; i < W; i++) PR"%c", lab1[bin[i]]);
  PR "|\n");
}

void plotFun(Fmin,Fmax,F,fmt)
     float Fmin,Fmax; double (*F)(); char *fmt;
/*===========================================*/
/* Assumes that plotInit() was called before		    */
/* Fmin, Fmax, F()  are in internal (unscaled by CX) units  */
{
  register short i;
  register short ind;
  short n[LINES];
  register float *end;
  register float *ptr;
  register float x;
  register float y, Ystep;
  float *dots[LINES], store[MAXCOLUMN];
  /* all x-values are stored in store[];
     end  points to the first unoccupied element of  store[];
     x-values corresponding to the same y-value have to be put into
     the adjacent elements of  store[]
  */

  end = store;
  for(i = 0; i < LINES; i++) n[i] = 0;

  Ystep = (Fmax - Fmin) / LINES;
  for(x = X1 + Xstep * 0.5; x < X2; x += Xstep)
    { y = (*F)(x);
      ind = (y - Fmin) / Ystep;
      if(ind >= 0 && ind < LINES)
	{ if(ind == LINES - 1) { if(n[ind] == 0) dots[ind] = end;
				 *end = x;
				 end++;
			       }
	    else 
	      { i = ind + 1;
		while(n[i] == 0 && i < LINES) i++;
		if(i == LINES) { *end = x;
				 if(n[ind] == 0) dots[ind] = end;
				 end++;
			       }
			else  { /* set new pointer: */
				 if(n[ind] == 0) dots[ind] = dots[i];
				/* shift x-values: */
				 for(ptr = end; ptr > dots[i]; --ptr)
							*ptr = *(ptr - 1);
				/* put in the new x-value: */
				 *dots[i] = x;
				/* shift pointers: */
				 end++; 
			         while(i < LINES) { if(n[i] != 0) (dots[i])++;
						    i++;
						  }
			      }
	      }
	  ++(n[ind]);
	}
    }
  ind = LINES;
  for(y = Fmax - Ystep * 0.5; y > Fmin; y -= Ystep)
		{ --ind; plotF(CY * y, n[ind], dots[ind], fmt); }
  xRange();
  (void)fflush(Listing);
}


void plotFun1(F,fmt)  double (*F)(); char *fmt;
/*===========================================*/
/* Calculats min and max of F(), then calls plotFun  */
{
  float x, y, Fmin  = 1e30, Fmax = -1e30;

  for(x = X1 + Xstep * 0.5; x < X2; x += Xstep)
	{ y = (*F)(x);
	  if(y > Fmax) Fmax = y;
	  if(y < Fmin) Fmin = y;
	}
  y = (Fmax - Fmin) * 0.02;
  Fmax += y;
  Fmin -= y;
  plotFun(Fmin,Fmax,F,fmt);
}


void Histogram(dens, N_x, xm, xs)
     float dens[]; int N_x; double xm, xs;
/*======================================*/
/* Assumes that plotInit() was called before
   dens[] contains N_x values corresponding to the x-coordinates
   equal to x_i = xm + xs/2 + i * xs; i = 0,1,...,N_x-1
*/
{
/*  extern unsigned sizeof(); */
  static char *fmt = "%.8f";
  register short ix, i, ind;
  short n[LINES];
  register float x, y, *store, *end, *ptr;
  float ymax = 0., ymin, Ystep, *dots[LINES];
  /* all x-values are stored in store[];
     end  points to the first unoccupied element of  store[];
     x-values corresponding to the same y-value have to be put into
     the adjacent elements of  store[]
  */

  /* Allocate memory for N_x elements in store[]  */
  if(( store = (float *) malloc((unsigned) (N_x * sizeof(float))) ) == NULL)
	{ PR"malloc in  Histogram (LinePlots.c) failed!\n\n");
	  exit(1);
	}

  end = store;
  for(i = 0; i < LINES; i++) n[i] = 0;

  for(ix = 0; ix < N_x; ix++) { if(dens[ix] > ymax) ymax = dens[ix]; }
  i = (ymax + 0.1) * 10.;
  ymax = (float)i / 10.;
  Ystep = ymax / (LINES - 1);
  ymin = -Ystep * 0.5;
  x = xm + 0.5 * xs;
  for(ix = 0; ix < N_x; ix++) 
    { y = dens[ix];
      ind = (y - ymin) / Ystep;
      if(ind == LINES - 1) { if(n[ind] == 0) dots[ind] = end;
			     *end = x;
			     end++;
			   }
	    else 
	      { i = ind + 1;
		while(n[i] == 0 && i < LINES) i++;
		if(i == LINES) { *end = x;
				 if(n[ind] == 0) dots[ind] = end;
				 end++;
			       }
			else  { /* set new pointer: */
				 if(n[ind] == 0) dots[ind] = dots[i];
				/* shift x-values: */
				 for(ptr = end; ptr > dots[i]; --ptr)
							*ptr = *(ptr - 1);
				/* put in the new x-value: */
				 *dots[i] = x;
				/* shift pointers: */
				 end++; 
			         while(i < LINES) { if(n[i] != 0) (dots[i])++;
						    i++;
						  }
			      }
	      }
	  ++(n[ind]);
	  x += xs;
    }
  ind = LINES;
  for(y = ymax; y > ymin; y -= Ystep)
		{ --ind; plotF(y, n[ind], dots[ind], fmt); }
  xRange();
  (void)fflush(Listing);
#ifdef ultrix
  (void)free((void *)store);
#else
  (void)free((char *)store);
#endif
}


