/* --------------- Adapted from "Numerical recipes in C" --------------- */

#include <stdio.h>
#include <math.h>

#define PE	(void)fprintf(stderr,
#define PR	(void)fprintf(Listing,

extern double EPSromb, xMinAt, xMaxAt, LAt;  /* LAt = (xMaxAt - xMinAt) */
					     /* xMinAt = a, xMaxAt = b  */
extern int JMAXromb, Kromb;
extern FILE *Listing;

static double *s, *h, *c, *d, a, b, b_a;

void qromb_init()
/*=============*/
{
/*
#ifdef ultrix
	extern void *malloc();
#else
	extern char *malloc();
#endif
*/

	s = (double *)malloc( (unsigned)(JMAXromb+1) * sizeof(double) ) - 1;
	h = (double *)malloc( (unsigned)(JMAXromb+1) * sizeof(double) ) - 1;
	c = (double *)malloc( (unsigned)Kromb * sizeof(double) ) - 1;
	d = (double *)malloc( (unsigned)Kromb * sizeof(double) ) - 1;
	a = xMinAt;
	b = xMaxAt;
	b_a = LAt;
	if(JMAXromb < Kromb)
	 { JMAXromb = Kromb;
	   PR"\nQROMB_INIT: JMAXromb increased to be equal to Kromb = %d\n\n",
					 Kromb);
	 }
}

void nrerror()
/*==========*/
{
	void exit();

	PR"\t ... exiting ...\n\n");
	exit(1);
}

double qromb(func)  double (*func)();
/*=================================*/
{
	double ss,dss,trapzd();
	int j;
	void polint();

	h[1]=1.0;
	for (j=1;j<=JMAXromb;j++) {
		s[j]=trapzd(func,j);
		/* PE"j = %d, s = %.16g\n", j, s[j]); */
		if (j >= Kromb) {
			polint(&h[j-Kromb],&s[j-Kromb],&ss,&dss);
			/* PE"ss = %.16g, dss = %.16g\n", ss, dss); */
			if (fabs(dss) < EPSromb*fabs(ss)) return ss;
		}
		s[j+1]=s[j];
		h[j+1]=0.25*h[j];
	}
	PE"\nQROMB: Convergence not achieved\n");
	PR
	"\nQROMB: Convergence not achieved with EPS = %g, JMAX = %d, K = %d\n",
						EPSromb, JMAXromb, Kromb);
	PR"ss = %g, error estimate dss = %g, EPS*fabs(ss) = %g\n",
						ss, dss, EPSromb*fabs(ss));
	nrerror();
}

#define FUNC(x) ((*func)(x))

double trapzd(func,n)  double (*func)(); int n;
/*===========================================*/
{
	double x;
	static double s, del;
	static int it;
	int j;

	if (n == 1) {
		it = 1;
		del = b_a;
		s = 0.5 * (FUNC(a) + FUNC(b));
		return (s * del);
	} else {
		x=a+0.5*del;
		for (j=1;j<=it;j++,x+=del) s += FUNC(x);
		it *= 2;
		del = b_a / it;
		return (s * del);
	}
}
#undef FUNC(x)

void polint(xa,ya,y,dy)  double xa[],ya[],*y,*dy;
/*=============================================*/
/* Neville's polynomial inter-/extrapolation from Kromb input points xa[i]
   to x = 0 */
{
	int i,m,ns=1;
	double den,dif,dift,ho,hp,w;

	dif=fabs(xa[1]);
	for (i=1;i<=Kromb;i++) {
		if ( (dift=fabs(xa[i])) < dif) {
			ns=i;
			dif=dift;
		}
		c[i]=ya[i];
		d[i]=ya[i];
	}
	*y=ya[ns--];
	for (m=1;m<Kromb;m++) {
	   for (i=1;i<=Kromb-m;i++) {
		ho=xa[i];
		hp=xa[i+m];
		w=c[i+1]-d[i];
		if((den=ho-hp) == 0.0)
		   { PR"\nPOLINT: two xa's are identical (within roundoff)\n");
		     nrerror();
		   }
		den=w/den;
		d[i]=hp*den;
		c[i]=ho*den;
	   }
	   *y += (*dy=(2*ns < (Kromb-m) ? c[ns+1] : d[ns--]));
	}
}
