/*************************************************************************/
/********************** (C) 1992 by Miroslav Kolar ***********************

   Generators of the pseudo/quasi-random numbers.

   Currently available: DEC's rand, random, erand48
			ranmar  (by George Marsaglia)
			halton
**************************************************************************/

/* Functions defined in this file:  ------------------------------------- 
double dummy()
double dbl_random()
double dbl_rand()
double erand()
void rmarin(ij, kl)  long int ij, kl;
double ranmar()
double halton()
double combHal()
double halton1d()
double halton1i()
void init_halton1R()
double halton1R()
void init_halton1R2b()
double halton1R2b()
double rdmHal2()
int init_rdmPhal2()
double rdmPhal2()
void Eratosthenes(primes,N) unsigned int *primes, N;
double (*init_state(iGen, rg, rgFlag, RnGn, State, Nfict))()
		char *rg; int iGen, *rgFlag, Nfict; union rgState *State;
void switch_gen(State, rgFlag)  union rgState *State; int rgFlag;
void refresh_rand_same(n)  int n;
void refresh_rand_diff(n)  int n;
int  init_rand_gen(rg1, rg2, actionFlag, Nfict) char **rg1, **rg2;
						int actionFlag, Nfict;
void describe_rand_gen()
void out_RnGn_defaults(Nfict) int Nfict;
void write_RnGn()
void read_RnGn()
void out_RnGn_state(mode,Listing) int mode; FILE *Listing;

-------------------------------------------------------------------------*/

/*  #define DEBUG  */     /* check switching between generators */
/*  #define DEBG1  */     /* for some random() related output only */

#include <values.h>

#include "1D.h"   /*  Contains  #include  and  #define  directives  */

#define SS		(void)sscanf(
#define HALTON_MAXBASE  40
#define MAX_D_HALTON_R  30	/* at most base^MAX_D_HALTON_R Halton numbers */
#define RANDOM_LEN	256 / sizeof(int)
#define min(x, y)  ((x) < (y) ? (x) : (y))

#define SINGLE_STREAM  -1
#define RANDOM		0
#define RAND		1
#define ERAND48		2
#define RANMAR		3
#define HALTON		4
#define HALTON1d	5
#define HALTON1i	6
#define HALTON1R	7
#define HALTON1R2b	8
#define RDMHAL2		9
#define COMBHAL		10
#define RDMPHAL2	11


/* Standard DEC's functions:
extern double erand48();
#ifdef ultrix
extern void srand();
#else
extern int srand();
#endif
extern int rand();
extern char *setstate(), *initstate();
extern long random();
*/

/* HIBITI and  BITS() defined in values.h */
static unsigned int HI_BIT = (1 << BITS(int) - 2);
					/* used for halton1R2b()	*/
static double dHIBITI,  /*   = (MAXINT + 1.);  see init_rand_gen();
		   used for halton1R2b()  and  dbl_random()
		   (dHIBITI = 2147483648.)      0 <= random() < dHIBITI	*/
	      maxrand;  /* see init_rand_gen(); used for dbl_rand()
			   0 <= rand() < maxrand			*/

/* Declared in 1D.c: ---------------------------------------------------- */
extern  FILE *Input, *Output, *Listing;
extern int rg1Flag, rg2Flag;
	  /* indicate the two selected generators for particle
             movement and configuration acceptance */

/*  extern double randNum1[], randNum2[];
           buffers to store N_fict random numbers generated by the
           two generators; these numbers are just enough for one MC cycle
*/
extern void err_fseek(), err_fread();
/* ---------------------------------------------------------------------- */

/* Other global declarations: ........................................... */
void  (*refresh_rand)();  /* refreshes the contents of randNum1[] and
			     randNum2[] using the selected generators  */
double (*RnGn1)(), (*RnGn2)(); /* pointers to the two selected generators */

#define RDM1 "ranmar(1874,8371)"
#define RDM2 SINGLE_STREAM_STRING
static char *DEFrandom[] = { "random(1,256)", "random(4123,256)" },
            *DEFrand[] = { "rand(1)", "rand(43123)" },
	    *DEFerand[] = { "erand48(1745,719,3)", "erand48(75,813,47345)" },
	    *DEFranmar[] = { RDM1, "ranmar(327,18534)" },
	    *DEFhalton[] = { "halton()() ... Nfict lowest pi; ni=0; pi=1",
			    "halton(1)(3)(0)(1)" },
	    *DEFhalton_class = "halton(1)(2)(0)(1)",
	    *DEFhalton1d[] = { "halton1d(2,0,1)", "halton1d(3,0,1)" },
	    *DEFhalton1i[] = { "halton1i(5,0,1)", "halton1i(7,0,1)" },
	    *DEFhalton1R[] = { "halton1R(3,111)", "halton1R(5,1719)" },
	    *DEFhalton1R2b[] = { "halton1R2b(1)", "halton1R2b(1719)" },
	    *DEFrdmHal2[] = { "rdm+hal2(1,1,256)", "rdm+hal2(1,4123,256)" },
	    *DEFrdmPhal2[] = { "rdmPhal2(1,256)", "rdmPhal2(4123,256)" },
	    *DEFcombHal[] = { "combHal()", "combHal(13)" };
char	    *DEFrg1 = RDM1,   /* To be used in       */
	    *DEFrg2 = RDM2;   /* defaults() in 1D.c  */
char	    *Single_stream_string = SINGLE_STREAM_STRING;

static int *last;  /* pointer to the last value returned by rand() */
static unsigned short *ss;  /* pointer to storage/seed space for erand48() */
static float *u, *c, *cd, *cm; /* state info for  ranmar() */
static long int *i97, *j97;    /*       - " -              */
static unsigned int *nBase, *iBase, *Base;  /* state info for  halton()   */
static double *nra1, *nrb1;		    /*       - " -                */
static unsigned int *nra1i, *nrb1i;	    /*       - " -     halton1i() */
static unsigned int *dig;	   	    /*       - " -     halton1R() */
static double *r;		  	    /*       - " -     halton1R() */
static unsigned int *invN;	   	    /*       - " -   halton1R2b() */
static int *state0,  /* pointer to state vector for random() - needed for
			a patch of a random() bug */
	   *periodP, /* pointer to period  - also needed for patch       */
	   rdmPhal2_shift1, rdmPhal2_shift2, *shift,
			/* used for rdmPhal2() - for shifting invN    */
	   *bytes;	/*        - " -        */ 

union rgState {				/* to store initialization and state */
 struct { unsigned seed;		/* vector of a RN generator	     */
	  int bytes;
	  int period;   /* used for patch */
	  int state[RANDOM_LEN];  } random;
 struct { unsigned seed;
	  int last;	    } rand;
 struct { unsigned short int ss0[3];
	  unsigned short int ss[3];  } erand48;
 struct { long int ij;
	  long int kl;
	  float c;
	  float cd;
	  float cm;
	  float u[97];
	  long int i97;
	  long int j97;     } ranmar;
 struct { unsigned int nBase;
	  unsigned int iBase;
	  unsigned int Base[HALTON_MAXBASE];
	  unsigned int nra0[HALTON_MAXBASE];
	  unsigned int nrb0[HALTON_MAXBASE];
	  double nra1[HALTON_MAXBASE];
	  double nrb1[HALTON_MAXBASE];		} halton;
 struct { unsigned int Base;
	  unsigned int nra0;
	  unsigned int nrb0;
	  double nra1;
	  double nrb1;		    } halton1d;
 struct { unsigned int Base;
	  unsigned int nra0;
	  unsigned int nrb0;
	  unsigned int nra1;
	  unsigned int nrb1;	    } halton1i;
 struct { unsigned int Base;
	  unsigned int n0;
	  unsigned int dig[MAX_D_HALTON_R + 1];
	  double r[MAX_D_HALTON_R + 1];		} halton1R;
 struct { unsigned int n0;
	  unsigned int invN;			} halton1R2b;
 struct { unsigned seed;
	  int bytes;
	  int period;   /* used for random() patch */
	  int state[RANDOM_LEN];
	  unsigned int n0;
	  unsigned int invN;			} rdmHal2
};
static union rgState rg1_state, rg2_state;

/* ...................................................................... */



double dummy()
/*==========*/
{ static int i = 0;
  PR"A dummy generator selected!\n");
  return (double)(i++, i %= 23);
 }


double dbl_random()
/*===============*/
{
 return *state0 += 5, *state0 %= *periodP, (double)random() / dHIBITI;

/* Here state0 points to the first element of the currently used state
   vector (in the int form).
   The first two statements above are a patch to overcome the following
   deficiency of random(). If two states are initialized using
   initstate(), then when switching between the two states using setstate(),
   setstate changes the value of the first element of the OLD state vector
   in the above manner. 
   If one wants to save the state vectors in a file and continue the
   calculation sometime later with the same stream of pseudorandom
   numbers, standard initstate()/setstate() does not provide for this
   situation. [Even if there is a single state whose first element is normally
   not incremented in the above manner, after a write/read intermission,
   one gets wrong results! This means that the state vector alone doesn't
   carry a complete information about the current state of random(). Something
   else must be set when initstate() is called. Two identical state vectors
   can give different streams under different conditions, and on the other hand
   two different state vectors - apparently difference can be  in the first
   element only - can give identical streams.] The above patch then seems to
   substitute for the normal interplay of initstate()/setstate(), and works in
   both situations: one or two state vectors. After reading in previously
   saved states that were modified in the above manner, it is sufficient to
   perform setstate() on them, and one is back in the correct stream
   of pseudorandom numbers.
   
	bytes      initial value of *state0      period = *periodP
	   8			0			....
	  32			1			 35
	  64			2			 75
	 128			3			155
	 256			4			315

   Actually, the order of operations is not critical here: one could first
   call random() and only then correct the value of *state0.
*/
}

double dbl_rand()
/*=============*/
{
 return (double)(*last = rand()) / maxrand;
}

double erand()
/*==========*/
{
 return erand48(ss);
}


/*----------------------------  RANMAR  ---------------------------------*/
/* This is the random number generator proposed by George Marsaglia in
   Florida State University Report: FSU-SCRI-87-50
   It was slightly modified by F. James to produce an array of pseudorandom
   numbers.

   From  the original Fortran version  ranmar.f  translated by f2c
   (version of 17 June 1991  16:43:01).
   Then manually modified by M. Kolar to remove all references to
   -lf2c   and  "f2c.h", and to simplify the code.
*/

/* Subroutine */
void rmarin(ij, kl)  long int ij, kl;
/*=================================*/
{
    extern void exit();
    /* Local variables */
    static long int i, j, k, l, m;
    static float s, t;
    static long int ii, jj;

/*
 This is the initialization routine for the random number generator RANMAR()
  NOTE: The seed variables can have values between:    0 <= ij <= 31328 
                                                       0 <= kl <= 30081 
 The random number sequences created by these two seeds are of sufficient
 length to complete an entire calculation with. For example, if several 
 different groups are working on different parts of the same calculation,
 each group could be assigned its own ij seed. This would leave each group
 with 30000 choices for the second seed. That is to say, this random
 number generator can create 900 million different subsequences -- with 
 each subsequence having a length of approximately 10^30.

 Use ij = 1802 & kl = 9373 to test the random number generator. The
 subroutine RANMAR should be used to generate 20000 random numbers.
 Then display the next six random numbers generated multiplied by 4096*4096
 If the random number generator is working properly, the random numbers 
 should be:
            6533892.0  14220222.0  7275067.0
            6172232.0  8354498.0   10633180.0
*/
    if (ij < 0 || ij > 31328 || kl < 0 || kl > 30081) {
	PR" The first random number seed must have a value between 0 and\
 31328");
	PR" The second seed must have a value between 0 and 30081");
	PR"trouble in rmarin");
	exit(-1);
    }
    i = ij / 177 % 177 + 2;
    j = ij % 177 + 2;
    k = kl / 169 % 178 + 1;
    l = kl % 169;
    for (ii = 1; ii <= 97; ++ii) {
	s = (float)0.;
	t = (float).5;
	for (jj = 1; jj <= 24; ++jj) {
	    m = i * j % 179 * k % 179;
	    i = j;
	    j = k;
	    k = m;
	    l = (l * 53 + 1) % 169;
	    if (l * m % 64 >= 32) {
		s += t;
	    }
	    t *= (float).5;
	}
	u[ii - 1] = s;
    }
    *c = (float).021602869033813477;
    *cd = (float).45623308420181274;
    *cm = (float).99999982118606567;
    *i97 = 97 - 1;
    *j97 = 33 - 1;
    return;
} /* rmarin_ */

double ranmar()
/*===========*/
{
    /* Local variables */
    register float uni;

    /* if(! test) {...}  No test for initialization made in this version!!! */
    uni = u[*i97] - u[*j97];
    if (uni < (float)0.) uni += (float)1.;
    u[*i97] = uni;
    --*i97;
    if (*i97 == -1) *i97 = 96;
    --*j97;
    if (*j97 == -1) *j97 = 96;
    *c -= *cd;
    if (*c < (float)0.) *c += *cm;
    uni -= *c;
    if (uni < (float)0.) uni += (float)1.;
    return (double)uni;
} /* ranmar_ */
/*--------------------  end of  RANMAR  ---------------------------------*/


/*----------------------------  HALTON  ---------------------------------*/
/* This is the generator of the quasi-random Halton numbers.
   Uses the algorithm by
        M. Berblinger & C. Schlier, Computer Phys. Commun. 66 (1991), 157.
   Converted to C by M. Kolar.

   nBase  bases  are used, subsequent calls to  halton()  use the bases
   periodically. That is, nBase calls to  halton()  generate one
   Halton vector.
   The bases should be mutually prime, as small as possible.
*/

#define BSHALTON base = Base[*iBase];\
	 nxa1 = nrb1[*iBase] - nra1[*iBase];\
	 if(nxa1 == (double)1.) {\
	    nra1[*iBase] = 1;\
	    nrb1[*iBase] *= base;\
	 }\
	 else\
	 {  nxb1 = nrb1[*iBase] / base;\
	    while(nxa1 <= nxb1) nxb1 /= base;\
	    nra1[*iBase] = (1 + base) * nxb1 - nxa1;\
	 }\
	 nxb1 = nra1[*iBase] / nrb1[*iBase];\

double halton()
/*===========*/
/* nra1[] and nrb1[] always contain integers (see the above paper).
   Here we use double precision to increase the maximum length
   of the generated sequence beyond 2^31/max(Base[])
*/
{
 register unsigned int base;
 register double nxb1;
 register double nxa1;

 BSHALTON
 ++*iBase;  *iBase %= *nBase;
 return nxb1;
}
/*--------------------  end of  HALTON  ---------------------------------*/

/*----------------------------  COMBHAL  --------------------------------*/
/* All the components of a Halton vector combined according to P. L'Ecuyer
   to give a single (pseudo-random???) number.
   The calculation of the individual components is the same as above.  
*/

double combHal()
/*============*/
{
 register unsigned int base;
 register double nxb1;
 register double nxa1;
 register double comb;

 for(*iBase = 0, comb = 0; *iBase < *nBase; ++*iBase) 
	{ BSHALTON
	  comb += nxb1;
	}
 while(comb > 1) comb -= 1;
 return comb;

}
/*--------------------  end of  COMBHAL  --------------------------------*/

/*----------------------------  HALTON1d  -------------------------------*/
/* This is the generator of the quasi-random Halton numbers using
   a SINGLE base only.  Otherwise the same as halton() above.
   DOUBLE ARITHMETICS.
*/

double halton1d()
/*=============*/
/* nra1 and nrb1 always contain integers (see the above paper).
   Here we use double precision to increase the maximum length
   of the generated sequence beyond 2^31/max(Base[])
*/
{
 register double nxb1;
 register double nxa1;

 nxa1 = *nrb1 - *nra1;
 if(nxa1 == (double)1.) {
    *nra1 = 1.;
    *nrb1 *= *Base;
 }
 else
 {  nxb1 = *nrb1 / *Base;
    while(nxa1 <= nxb1) nxb1 /= *Base;
    *nra1 = (1. + *Base) * nxb1 - nxa1;
 }
 return *nra1 / *nrb1;
}
/*--------------------  end of  HALTON1d  --------------------------------*/

/*----------------------------  HALTON1i  --------------------------------*/
/* This is the generator of the quasi-random Halton numbers using
   a SINGLE base only.  Otherwise the same as halton() above.
   INTEGER ARITHMETICS.
*/

double halton1i()
/*=============*/
{
 register unsigned int nxb1;
 register unsigned int nxa1;

 nxa1 = *nrb1i - *nra1i;
 if(nxa1 == 1) {
    *nra1i = 1;
    *nrb1i *= *Base;
 }
 else
 {  nxb1 = *nrb1i / *Base;
    while(nxa1 <= nxb1) nxb1 /= *Base;
    *nra1i = (1 + *Base) * nxb1 - nxa1;
 }
 return (double)*nra1i / (double)*nrb1i;
}
/*--------------------  end of  HALTON1i  --------------------------------*/

/*----------------------------  HALTON1R  --------------------------------*/
/* This is the generator of the quasi-random Halton numbers using
   a SINGLE base only and the original algorithm: reversal of the digits
   in the in the number system with the given base.
   See: M. Kolar & S.F. Oshea: Fast, Portable, and Reliable Algorithm for
	the Calculation of Halton Numbers
   See also the file  RanGen1.c
*/

void init_halton1R(n0) unsigned n0;
/*===============================*/
{
 unsigned int last, i;

 /* The generator will start with n0-th Halton number. Thus it must be set
    to the state that corresponds to the successfully finished calculation
    of the (n0-1)-th Halton number: */
 if(n0 < 1) n0 = 1;
 --n0;
 last = 0;
 while(n0 >= *Base) { dig[last++] = n0 % *Base; n0 /= *Base; }
 dig[last] = n0;
 for(i = last + 1; i < MAX_D_HALTON_R; i++) dig[i] = 0;
 r[MAX_D_HALTON_R] = 0;
 for(i = MAX_D_HALTON_R; i > 0; i--) r[i - 1] = (dig[i] + r[i]) / *Base;
}

double halton1R()
/*=============*/
{
/* register unsigned short last = 0;
 register unsigned short i;		rather faster  is:
*/ register unsigned int last = 0;
 register unsigned int i;

 if(++dig[0] == *Base)
    { do dig[last++] = 0; while(++dig[last] == *Base);
      r[last - 1] = (dig[last] + r[last]) / *Base;
      for(i = last - 1; i > 0 ; i--) r[i - 1] = r[i] / *Base;
      return r[0] / *Base;
    }
 else return (dig[0] + r[0]) / *Base;
}
/*--------------------  end of  HALTON1R  --------------------------------*/

/*---------------------------  HALTON1R2b --------------------------------*/
/* This is the generator of the quasi-random Halton numbers using
   a SINGLE base 2 only and the original algorithm: reversal of the digits
   in the in the binary number system using BIT operations.
   See: M. Kolar & S.F. Oshea: Fast, Portable, and Reliable Algorithm for
	the Calculation of Halton Numbers
   See also the file  RanGen1.c
*/

void init_halton1R2b(n0) unsigned n0;
/*=================================*/
{
 int b;

 /* The generator will start with n0-th Halton number. Thus it must be set
    to the state that corresponds to the successfully finished calculation
    of the (n0-1)-th Halton number: */
 if(n0 < 1) n0 = 1;
 --n0;

 /* To invert the order of bits, first transfer the lowest bit from n0 to invN.
    Then repeatedly shift n0 by one bit to the right, invN by one bit
    to the left, and every time transfer the lowest bit from n0 to invN:
 */
 *invN = n0 & 01;
 for(b = 1; b < BITS(int) - 1; b++)
	 { *invN <<= 1; *invN |= ((n0 >>= 1) & 01); }
}

double halton1R2b()
/*===============*/
{
 register unsigned int b;

 if(*invN & HI_BIT)
    { /* at least one carry needed */
      *invN &= ~HI_BIT;
      b = HI_BIT >> 1;
      while(*invN & b) { *invN &= ~b; b >>= 1; }
      *invN |= b;
    }
 else *invN |= HI_BIT;

 return *invN / dHIBITI;
}
/*-------------------  end of  HALTON1R2b  -------------------------------*/


/*-----------------------------  RDMHAL2  --------------------------------*/
/* This is the combination of  halton1R2b()  +  random()
   in the sense of Lemma 1 of
   P. L'Ecuyer, Communications of the ACM, 31 (6), 742-750 (1988)

   Improves  random()  considerably!!!!!!!
 
   [For single stream, similar combination of arbitrary two generators can
   be achieved by  SelectMode = PC  in the main program (command line)]   */

double rdmHal2()
/*============*/
{
 register unsigned int b;
 double res;

 if(*invN & HI_BIT)
    { /* at least one carry needed */
      *invN &= ~HI_BIT;
      b = HI_BIT >> 1;
      while(*invN & b) { *invN &= ~b; b >>= 1; }
      *invN |= b;
    }
 else *invN |= HI_BIT;

 *state0 += 5;
 *state0 %= *periodP;

 res = (double)*invN + (double)random();
 if(res >= dHIBITI) res -= dHIBITI;
 return res / dHIBITI;

}
/*--------------------------  end of RDMHAL2 ----------------------------*/


/*-----------------------------  RDMPHAL2  --------------------------------*/
/* This is an attempt to improve  random()
   by changing  i-th  element of (int *)state, where  0 < i < RANDOM_LEN
   is generated using halton1R2b()  -  invN .

   !!!!!!!!!!!!!!! IT SEEMS TO BE A BIG NONSENSE !!!!!!!!!!!!!!!!!!
                   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

int init_rdmPhal2()
/*===============*/
{
 int ratio, shift = 0;

 ratio = dHIBITI / *bytes * sizeof(int);
 while(ratio > 1) { ratio /= 2; shift++; }
 return shift;
}

double rdmPhal2()
/*=============*/
{
 register unsigned int b;
 double res;
 int ind;

 *state0 += 5;
 *state0 %= *periodP;
 res = (double)random() / dHIBITI;

 if(*invN & HI_BIT)
    { /* at least one carry needed */
      *invN &= ~HI_BIT;
      b = HI_BIT >> 1;
      while(*invN & b) { *invN &= ~b; b >>= 1; }
      *invN |= b;
    }
 else *invN |= HI_BIT;

 ind = *invN >> *shift;
 if(ind) *(state0 + ind) = ~ *(state0 + ind);
		/* This  seems to be worse than random() itself!!!!	*/
/* if(ind) *(state0 + ind) |= *invN;  This is even worse!!!!		*/
/* if(ind) *(state0 + ind) &= *invN;  This is still worse!!!!		*/

 return res;

}
/*--------------------------  end of RDMPHAL2 ----------------------------*/


void Eratosthenes(primes,N)  unsigned int *primes, N;
/*=================================================*/
/* Sieve of Eratosthenes */
{
  int np, x, k, p, j, flg = primes[0] <= 0;

/* If the following 5 lines: */
  np = 0;
  x = 2.;
  while((p = primes[np]) > 0) { if(p > x) x = p; if(++np == N) return; }
  if(np == 0) { primes[0] = 2; np++; }
  x++;
/* were substituted with the next 2 lines:
   primes[0] = 2;
   np = 1; x = 3;
and below only the branch corresponding to (flg != 0) used, this function would
generate the first N primes and put them into the array primes[], regardless
of its previous contents.
In the present version, if initially primes[0] <= 0, one gets the same result.
Otherwise, the array primes[] is completed with the lowest integers that have
no common divisors with all the positive integers already stored at the
beginning of primes[] and larger then all such previous primes[].
*/
  if(flg)
     while(np < N)
	{ k = sqrt((double)x) + 1e-5;
	  for (j = 0; (p = primes[j]) <= k; j++) { if (x % p == 0) break; }
	  if(primes[j] > k) primes[np++] = x;
	  x++;
	}
  else
     while(np < N)
	{ for (j = 0; j < np; j++)
		{ p = primes[j];
		  for (k = 2; k <= p; k++)
			 { if (p % k == 0 && x % k == 0) break; }
		  if(k <= p) break;
		}
	  if(j == np) primes[np++] = x;
	  x++;
	}

}  /* ----------------  end of:  Eratosthenes()  ------------------- */


/* -----------------------  init_state():  ----------------------------- */

#define CHECK(N) { ch = rg + N; while( *ch != '\0')\
		if( isdigit((int)*ch) || *ch == ',' || *ch == '(' || *ch == ')'\
		    || ((*rgFlag == HALTON || *rgFlag == COMBHAL) &&\
		    !strncmp(ch,"rand",4) && (ch += 3, 1)) ) ++ch;\
		else { PE"\nIncorrect RNG specification: r%d%s\n\n",\
			   iGen + 1, rg);  return(NULL); } }

#define GET_HALT(pntr,Lbl) { while(*ch != '\0' && *ch++ != '(') ;\
			     if(*ch == '\0') goto Lbl;\
			     if(*ch != ')') for (i = 0; i < *nBase ; i++)\
			      { if(sscanf(ch, "%u", pntr + i) != 1) break;\
				while(isdigit((int)*ch)) ch++;\
				if(*ch != ',') break; ch++; } }

#define RAND_INIT { for(i = 0; i < *nBase; i++) sssd += Base[i];\
		    srand(sssd);\
		    for(i = 0; i < *nBase; i++)\
		     { nra[i] = (rand() / maxrand) * 50000;\
		       nrb[i] = (rand() / maxrand) * 50000; } }

double (*init_state(iGen, rg, rgFlag, State, Nfict))()
       char *rg; int iGen, *rgFlag, Nfict; union rgState *State;
/*============================================================*/
{
 int i, halFlg;
 char *ch;
#ifdef ultrix
 unsigned sssd = 1737;
#else
 int sssd = 1737;
#endif

 /* iGen = 1,2  ...  enumerates the two selected generators */
 iGen--;

 if(! strncmp(rg,"random",6))
      { *rgFlag = RANDOM;
	CHECK(6)
	/* first fill with default values, then check whether some values
	   redefined on the command line */
	periodP = &State->random.period;
	state0 = State->random.state;
	SS DEFrandom[iGen] + 6, "(%u,%d)", & State->random.seed,
					   & State->random.bytes);
	SS rg + 6, "(%u,%d)", & State->random.seed, & State->random.bytes);
	if(State->random.bytes < 32) { State->random.bytes = 8; *periodP = 1; }
	else if(State->random.bytes < 64) { State->random.bytes = 32;
					    *periodP = 35; }
	else if(State->random.bytes < 128) { State->random.bytes = 64;
					     *periodP = 75; }
	else if(State->random.bytes < 256) { State->random.bytes = 128;
					     *periodP = 155; }
	else { State->random.bytes = 256; *periodP = 315; }

	(void)initstate(State->random.seed, (char *)state0,
			State->random.bytes);
	return(dbl_random);
      }

 else if(! strncmp(rg,"rand",4))
      { *rgFlag = RAND;
	CHECK(4)
	/* first fill with default values, then ... */
	last = & State->rand.last;
	SS DEFrand[iGen] + 4, "(%u)", last);
	SS rg + 4, "(%u)", last);
	State->rand.seed = (unsigned)*last;
#ifdef ultrix
	srand((unsigned)(*last));
#else
	srand((*last));
#endif
	return(dbl_rand);
      }

 else if(! strncmp(rg,"erand48",7))
      { *rgFlag = ERAND48;
	CHECK(7)
	ss = State->erand48.ss;
	/* first fill with default values, then ... */
	SS DEFerand[iGen] + 7, "(%hu,%hu,%hu)", ss, ss + 1, ss + 2);
	SS rg + 7, "(%hu,%hu,%hu)", ss, ss + 1, ss + 2);
	State->erand48.ss0[0] = ss[0];
	State->erand48.ss0[1] = ss[1];
	State->erand48.ss0[2] = ss[2];
	return(erand);
      }

 else if(! strncmp(rg,"ranmar",6))
      { long int ij, kl;

	*rgFlag = RANMAR;
	CHECK(6)
	/* first fill with default values, then ... */
	SS DEFranmar[iGen] + 6, "(%ld,%ld)", &ij, &kl);
	SS rg + 6, "(%ld,%ld)",  &ij, &kl);
	State->ranmar.ij = ij;
	State->ranmar.kl = kl;
	u = State->ranmar.u;
	c = & State->ranmar.c;
	cd = & State->ranmar.cd;
	cm = & State->ranmar.cm;
	i97 = & State->ranmar.i97;
	j97 = & State->ranmar.j97;
	rmarin(ij, kl);
	return(ranmar);
      }

 else if(! strncmp(rg,"halton1R2b",10))
      { unsigned int *n0;

	*rgFlag = HALTON1R2b;
	CHECK(10)
	n0 = & State->halton1R2b.n0;
	invN = & State->halton1R2b.invN;
	/* first fill with default values, then check whether some values
	   redefined on the command line */
	*n0 = 1;
	SS DEFhalton1R2b[iGen] + 10, "(%u)", n0);
	SS rg + 10, "(%u)", n0);
	init_halton1R2b(*n0);
	return(halton1R2b);
      }

 else if(! strncmp(rg,"rdm+hal2",8))
      { unsigned int *n0;

	*rgFlag = RDMHAL2;
	CHECK(8)
	/* first fill with default values, then check whether some values
	   redefined on the command line */
	periodP = &State->rdmHal2.period;
	state0 = State->rdmHal2.state;
        bytes = &State->rdmHal2.bytes;
	n0 = & State->rdmHal2.n0;
	invN = & State->rdmHal2.invN;
	*n0 = 1;
	SS DEFrdmHal2[iGen] + 8, "(%u,%u,%d)", n0, &State->rdmHal2.seed, bytes);
	SS rg + 8, "(%u,%u,%d)", n0, & State->rdmHal2.seed, bytes);
	if(*bytes < 32) { *bytes = 8; *periodP = 1; }
	else if(*bytes < 64) { *bytes = 32; *periodP = 35; }
	else if(*bytes < 128) { *bytes = 64; *periodP = 75; }
	else if(*bytes < 256) { *bytes = 128; *periodP = 155; }
	else { *bytes = 256; *periodP = 315; }
	(void)initstate(State->rdmHal2.seed, (char *)state0, *bytes);
	init_halton1R2b(*n0);
	return(rdmHal2);
      }

 else if(! strncmp(rg,"rdmPhal2",8))
      { *rgFlag = RDMPHAL2;
	CHECK(8)
	/* first fill with default values, then check whether some values
	   redefined on the command line */
	periodP = &State->rdmHal2.period;
	state0 = State->rdmHal2.state;
        bytes = &State->rdmHal2.bytes;
	State->rdmHal2.n0 = 1;
	invN = & State->rdmHal2.invN;
	SS DEFrdmPhal2[iGen] + 8, "(%u,%d)", &State->rdmHal2.seed, bytes);
	SS rg + 8, "(%u,%d)", & State->rdmHal2.seed, bytes);
	if(*bytes < 32) { *bytes = 8; *periodP = 1; }
	else if(*bytes < 64) { *bytes = 32; *periodP = 35; }
	else if(*bytes < 128) { *bytes = 64; *periodP = 75; }
	else if(*bytes < 256) { *bytes = 128; *periodP = 155; }
	else { *bytes = 256; *periodP = 315; }
	(void)initstate(State->rdmHal2.seed, (char *)state0, *bytes);
	init_halton1R2b(1);
	shift = iGen ? &rdmPhal2_shift2 : &rdmPhal2_shift1;
	*shift = init_rdmPhal2();
	return(rdmPhal2);
      }

 else if(! strncmp(rg,"halton1R",8))
      { unsigned int *n0;

	*rgFlag = HALTON1R;
	CHECK(8)
	Base = & State->halton1R.Base;
	n0 = & State->halton1R.n0;
	dig = State->halton1R.dig;
	r = State->halton1R.r;
	/* first fill with default values, then check whether some values
	   redefined on the command line */
	*n0 = 1;
	SS DEFhalton1R[iGen] + 8, "(%u,%u)", Base, n0);
	SS rg + 8, "(%u,%u)", Base, n0);
	init_halton1R(*n0);
	return(halton1R);
      }

 else if(! strncmp(rg,"halton1i",8))
      { unsigned int *nra, *nrb;

	*rgFlag = HALTON1i;
	CHECK(8)
	Base = & State->halton1i.Base;
	nra1i = & State->halton1i.nra1;
	nrb1i = & State->halton1i.nrb1;
	nra = & State->halton1i.nra0;
	nrb = & State->halton1i.nrb0;
	/* first fill with default values, then check whether some values
	   redefined on the command line */
	SS DEFhalton1i[iGen] + 8, "(%u,%u,%u)", Base, nra, nrb);
	SS rg + 8, "(%u,%u,%u)", Base, nra, nrb);

	/* check the compatibility of nra0 and nrb0: */
		  if(*nra == 0 || *nrb == 0) *nrb = 1;
		  /* if(*nrb < *Base ) *nrb = 1;
			 else *nrb = (*nrb / *Base ) * *Base ; */
		  while(*nra > *nrb )
			if(*nrb >= dHIBITI / *Base ) *nra = 0.9 * *nrb; 
			  else  *nrb *= *Base ;
		  while(*nra % *Base == 0 && *nrb % *Base == 0)
					 { *nra /= *Base ;
					   *nrb /= *Base ;
					 }
		  *nra1i  = *nra ;
		  *nrb1i  = *nrb ;
	return(halton1i);
      }

 else if(! strncmp(rg,"halton1d",8))
      { unsigned int *nra, *nrb;

	*rgFlag = HALTON1d;
	CHECK(8)
	Base = & State->halton1d.Base;
	nra1 = & State->halton1d.nra1;
	nrb1 = & State->halton1d.nrb1;
	nra = & State->halton1d.nra0;
	nrb = & State->halton1d.nrb0;
	/* first fill with default values, then check whether some values
	   redefined on the command line */
	SS DEFhalton1d[iGen] + 8, "(%u,%u,%u)", Base, nra, nrb);
	SS rg + 8, "(%u,%u,%u)", Base, nra, nrb);

	/* check the compatibility of nra0 and nrb0: */
		  if(*nra == 0 || *nrb == 0) *nrb = 1;
		  /* if(*nrb < *Base ) *nrb = 1;
			 else *nrb = (*nrb / *Base ) * *Base ; */
		  while(*nra > *nrb )
			if(*nrb >= dHIBITI / *Base ) *nra = 0.9 * *nrb; 
			  else  *nrb *= *Base ;
		  while(*nra % *Base == 0 && *nrb % *Base == 0)
					 { *nra /= *Base ;
					   *nrb /= *Base ;
					 }
		  *nra1  = (double)*nra ;
		  *nrb1  = (double)*nrb ;
	return(halton1d);
      }

 else if((halFlg = ! strncmp(rg,"halton",6)) || ! strncmp(rg,"combHal",7))
      { unsigned int *nra, *nrb, b;

	if(halFlg) { *rgFlag = HALTON; CHECK(6) }
	     else  { *rgFlag = COMBHAL; CHECK(7) }
	nBase = & State->halton.nBase;
	iBase = & State->halton.iBase;
	Base = State->halton.Base;
	nra1 = State->halton.nra1;
	nrb1 = State->halton.nrb1;
	nra = State->halton.nra0;
	nrb = State->halton.nrb0;
	/* here, the procedure is more complicated: at first get (implicit)
	   defaults from DEFhalton[], then everything what is available from
	   the command line, and in the end the explicit defaults are added
	   where necessary */

	*iBase = 0;
	for(i = 0; i < HALTON_MAXBASE; i++) Base[i] = nra[i] = nrb[i] = 0;	

	if(halFlg) {if(Nfict > 1) ch = DEFhalton[iGen] + 6;
			else  ch = (iGen ? DEFhalton[1] : DEFhalton_class) + 6;
		    /* Here, the value of Nfict is used to distinguish between
		       the quantum and classical versions of the program.
		       In the quantum version, Nfict is always greater or
		       equal than 2. */
		   }
	else ch = DEFcombHal[iGen] + 7;

	if(sscanf(ch++, "(%u", nBase) != 1) *nBase = Nfict;
	if(*nBase > HALTON_MAXBASE)
	      { *nBase = HALTON_MAXBASE;
		PR"\nNumber of Halton bases from DEFhalton[%d] reduced to\
 HALTON_MAXBASE = %d\n\n", iGen, HALTON_MAXBASE);
	      }
	GET_HALT(Base, ComLine)
	GET_HALT(nra, ComLine)
	if(strncmp(ch,"rand",4)) GET_HALT(nrb, ComLine)
		else		 RAND_INIT

  ComLine:    /* read values from command line */
	ch = rg + (halFlg ? 6 : 7);
	if(sscanf(ch++, "(%u", &b) == 1) *nBase = b;
	if(*nBase > HALTON_MAXBASE)
	      { *nBase = HALTON_MAXBASE;
		PR"\nNumber of Halton bases for the RN generator %d reduced\
 to HALTON_MAXBASE = %d\n\n", iGen + 1, HALTON_MAXBASE);
	      }
	GET_HALT(Base, Default)
	GET_HALT(nra, Default)
	if(strncmp(ch,"rand",4)) GET_HALT(nrb, Default)
		else	     RAND_INIT

  Default:   /* substitute explicite defaults for any missing values */
	Eratosthenes(Base, *nBase);
	for(i = 0; i < *nBase; i++)
		{
		  if(nra[i] == 0 || nrb[i] == 0) nrb[i] = 1;
		  /* if(nrb[i] < Base[i]) nrb[i] = 1;
			 else nrb[i] = (nrb[i] / Base[i]) * Base[i]; */
		  while(nra[i] > nrb[i])
			if(nrb[i] >= dHIBITI / Base[i]) nra[i] = 0.9 * nrb[i]; 
			  else  nrb[i] *= Base[i];
		  while(nra[i] % Base[i] == 0 &&
			nrb[i] % Base[i] == 0) { nra[i] /= Base[i];
						 nrb[i] /= Base[i];
					       }
		  nra1[i] = (double)nra[i];
		  nrb1[i] = (double)nrb[i];
		}

	return(halFlg ? halton : combHal);
      }

 else if(iGen && ! strcmp(rg, SINGLE_STREAM_STRING))
	{ *rgFlag = SINGLE_STREAM; return(dummy); }

 else { PE"\nUnknown random number generator specified: r%d%s\n\n",
	  iGen + 1, rg);
	return(NULL);
      }

} /* ------------------  end of:  init_state()  ----------------------- */


void switch_gen(State, rgFlag)  union rgState *State; int rgFlag;
/*=============================================================*/
{
 switch(rgFlag)
  { case RANMAR:  u = State->ranmar.u;
		  c = & State->ranmar.c;
		  cd = & State->ranmar.cd;
		  cm = & State->ranmar.cm;
		  i97 = & State->ranmar.i97;
		  j97 = & State->ranmar.j97;
		  break;
    case ERAND48: ss = State->erand48.ss;
		  break;
    case RAND:    last = & State->rand.last;
#ifdef ultrix
		  srand((unsigned)(*last));
#else
		  srand((*last));
#endif
		  break;
    case RANDOM:  state0 = State->random.state;
		  periodP = &State->random.period;
		  (void)setstate((char *)State->random.state);
 		  break;
    case HALTON1R2b: invN = & State->halton1R2b.invN;
		     break;
    case HALTON1R: Base = & State->halton1R.Base;
		   dig = State->halton1R.dig;
		   r = State->halton1R.r;
		   break;
    case HALTON1d: Base = & State->halton1d.Base;
		   nra1 = & State->halton1d.nra1;
		   nrb1 = & State->halton1d.nrb1;
		   break;
    case HALTON1i: Base = & State->halton1i.Base;
		   nra1i = & State->halton1i.nra1;
		   nrb1i = & State->halton1i.nrb1;
		   break;
    case COMBHAL:
    case HALTON:  nBase = & State->halton.nBase;
		  iBase = & State->halton.iBase;
		  Base = State->halton.Base;
		  nra1 = State->halton.nra1;
		  nrb1 = State->halton.nrb1;
		  break;
    case RDMPHAL2: shift =
		   (State == &rg1_state) ? &rdmPhal2_shift1 : &rdmPhal2_shift2;
    case RDMHAL2: state0 = State->rdmHal2.state;
		  periodP = &State->rdmHal2.period;
		  (void)setstate((char *)State->rdmHal2.state);
 		  invN = & State->rdmHal2.invN;
		  break;
  }
} /* ------------------  end of:  switch_gen()  ----------------------- */


void refresh_rand_same(n1, n2, randNum1, randNum2)
      int n1, n2;  double *randNum1, *randNum2;
/*==============================================*/
{
 register int i;

 switch_gen(&rg1_state, rg1Flag);
 for(i = 0; i < n1; i++) randNum1[i] = (*RnGn1)();
 switch_gen(&rg2_state, rg2Flag);
 for(i = 0; i < n2; i++) randNum2[i] = (*RnGn2)();

#ifdef DEBG1
 if(rg1Flag == RANDOM)
	PE"    %d %d", rg1_state.random.state[0], rg2_state.random.state[0]);

 if(n1 <= n2)
     { for(i = 0; i < n1; i++) PE"%.20g %.20g\n", randNum1[i], randNum2[i]);
       for(i = n1; i < n2; i++) PE"                      %.20g\n", randNum2[i]);
     }
     else
     { for(i = 0; i < n2; i++) PE"%.20g %.20g\n", randNum1[i], randNum2[i]);
       for(i = n2; i < n1; i++) PE"%.20g\n", randNum1[i]);
     }
 PE"\n");
#endif
} /* --- end of:  refresh_rand_same() */


void refresh_rand_diff(n1, n2, randNum1, randNum2)
     int n1, n2;  double *randNum1, *randNum2;
/*==============================================*/
{
 register int i;
 for(i = 0; i < n1; i++)  randNum1[i] = (*RnGn1)();
 for(i = 0; i < n2; i++)  randNum2[i] = (*RnGn2)();

#ifdef DEBG1
 if(rg1Flag == RANDOM) PE"    %d ", rg1_state.random.state[0]);
 if(rg2Flag == RANDOM) PE"    %d ", rg2_state.random.state[0]);
 if(rg1Flag == RANDOM || rg2Flag == RANDOM) PE"\n");

 if(n1 <= n2)
     { for(i = 0; i < n1; i++) PE"%.20g %.20g\n", randNum1[i], randNum2[i]);
       for(i = n1; i < n2; i++) PE"                      %.20g\n", randNum2[i]);
     }
     else
     { for(i = 0; i < n2; i++) PE"%.20g %.20g\n", randNum1[i], randNum2[i]);
       for(i = n2; i < n1; i++) PE"%.20g\n", randNum1[i]);
     }
 PE"\n");
#endif
} /* --- end of:  refresh_rand_diff() */


/* --------------------- init_rand_gen(): ------------------------ */

#define CHOOSE(rg,RnGn,State) switch(rg)\
{\
  case RANDOM:  RnGn = dbl_random;\
/*		periodP = &State.random.period;\
		state0 = State.random.state;*/\
		break;\
  case RAND:    RnGn = dbl_rand;\
		break;\
  case ERAND48: RnGn = erand;\
		break;\
  case RANMAR:  RnGn = ranmar;\
		break;\
  case HALTON:  RnGn = halton;\
		break;\
  case COMBHAL: RnGn = combHal;\
		break;\
  case RDMPHAL2: shift = (& State == &rg1_state) ?\
			  &rdmPhal2_shift1 : &rdmPhal2_shift2;\
		 bytes = &State.rdmHal2.bytes;\
		 *shift = init_rdmPhal2();\
		 RnGn = rdmPhal2;\
		 break;\
  case RDMHAL2: RnGn = rdmHal2;\
		break;\
  case HALTON1R2b: RnGn = halton1R2b;\
		   break;\
  case HALTON1R: RnGn = halton1R;\
		 break;\
  case HALTON1i: RnGn = halton1i;\
		 break;\
  case HALTON1d: RnGn = halton1d;\
		 break;\
  case SINGLE_STREAM: *rg2 = Single_stream_string;\
}
/* This Macro is used when the generator has already been initialized
   in a previous run. */


int init_rand_gen(rg1, rg2, actionFlag, Nfict)
     char **rg1, **rg2; int actionFlag, Nfict;
/*==========================================*/
{

dHIBITI = MAXINT + 1.;  /* must be initialized here because it is not
			   considered by the compiler to be a constant */
#ifdef RAND_MAX
maxrand = RAND_MAX + 1.;
#else
maxrand = dHIBITI; /* This is sort of gamble. It works for Ultrix and SPARC */
#endif


#ifdef DEBUG
PE" ------ Addresses (init_rand_gen - beginning) ------ \n");
PE"rg1: %p, rg2: %p, DEFrg1: %p, DEFrg2: %p\n", *rg1, *rg2, DEFrg1, DEFrg2);
PE"dbl_random: %p, dbl_rand: %p, erand: %p, ranmar: %p\n",
		dbl_random, dbl_rand, erand, ranmar);
PE"halton: %p, halton1d: %p, halton1i: %p, halton1R: %p\n",
		halton, halton1d, halton1i, halton1R);
PE"halton1R2b: %p, rdm+hal2: %p, combHal: %p, dummy: %p\n",
		halton1R2b, rdmHal2, combHal, dummy);
PE"refresh_rand_same: %p, refresh_rand_diff: %p\n",
		refresh_rand_same, refresh_rand_diff);
PE"RnGn1: %p, RnGn2: %p, refresh_rand: %p\n", RnGn1, RnGn2, refresh_rand);
#endif

 if(! actionFlag) { if(**rg1 == '\0') *rg1 = DEFrg1;
		    if(**rg2 == '\0') *rg2 = DEFrg2;
		  }
 /* At this point both **rg1 and **rg2 can be equal to '\0' if actionFlag > 0.
    Then the old generators from Input Master File are used, and below only
    the check whether the two generators are equal is done.
 */

 if(**rg1 != '\0')
     { if( (RnGn1 = init_state(1, *rg1, &rg1Flag, &rg1_state, Nfict)) == NULL )
							return(-1);
     }
     else  {  CHOOSE(rg1Flag, RnGn1, rg1_state)
	      switch_gen(&rg1_state, rg1Flag);
	   }
 if(**rg2 != '\0')
     { if( (RnGn2 = init_state(2, *rg2, &rg2Flag, &rg2_state, Nfict)) == NULL )
							return(-1);
     }
     else  {  CHOOSE(rg2Flag, RnGn2, rg2_state)
	      switch_gen(&rg2_state, rg2Flag);
	   }

 if(! strcmp(*rg2, SINGLE_STREAM_STRING)) RnGn2 = RnGn1;  /* This is needed
					for the moves of the whole polymer */

 if((rg1Flag == HALTON || rg1Flag == COMBHAL || rg2Flag == HALTON ||
			rg2Flag == COMBHAL ) && Nfict > HALTON_MAXBASE)
    { PR"\nTo accommodate %d Halton bases, increase HALTON_MAXBASE in\
 RanGen.c\n\n", Nfict);
      return(-1);
    }	 
 refresh_rand = (rg1Flag == rg2Flag) ? refresh_rand_same : refresh_rand_diff;

#ifdef DEBUG
PE" ------ Addresses (init_rand_gen - end) ------ \n");
PE"rg1: %p, rg2: %p\n", *rg1, *rg2);
PE"RnGn1: %p, RnGn2: %p, refresh_rand: %p\n", RnGn1, RnGn2, refresh_rand);
if(rg1Flag == ERAND48 || rg2Flag == ERAND48)
	 PE"ss: %p, rg1_state.erand48.ss: %p, rg2_state.erand48.ss: %p\n",
	   ss, rg1_state.erand48.ss, rg2_state.erand48.ss);
#endif

return(0);

} /* ------------------  end of:  init_rand_gen()  ----------------------- */


void describe_rand_gen()
/*====================*/
{
PE"  RandGen:  random(seed,bytes)  ... DEC's random, 8 <= bytes <= 256\n");
PE"            rand(seed)          ... DEC's rand\n");
PE"            erand48(s1,s2,s3)   ... DEC's erand48, 0 <= si < 65536\n");
PE"            ranmar(s1,s2)       ... Marsaglia, 0<=s1<=31328, 0<=s2<=30081\n");
PE"            halton(N)(p1,p2,...,pN)(n1,...)(d1,...)  ... N mutually prime\n");
PE"                                      bases pi; seeds ni/di: 0 <= ni < di\n");
PE"                                      (Berblinger & Schlier algorithm)\n");
/* PE"                 bases pi; seeds ni/di: di = pi**ki, 0 <= ni < di\n"); */
PE"            halton(N)(p1,p2,...,pN)(rand) ... random selection of seeds\n");
PE"                                             (Berblinger & Schlier algorithm)\n");
PE"            halton1d(p,n,d)     ... single base (p) Halton with seed n/d;\n");
PE"                                    double arithmetics (BS algorithm)\n");
PE"            halton1i(p,n,d)     ... single base (p) Halton with seed n/d;\n");
PE"                                    integer arithmetics (BS algorithm)\n");
PE"                                    -- slower than double!\n");
PE"            halton1R(p,n0)      ... single base (p) Halton with seed 0 but\n");
PE"                    starting with n0-th number - using digit reversal\n");
PE"            halton1R2b(n0)      ... as halton1R, but only for p = 2 using\n");
PE"                                    fast bit manipulations\n");
PE"            rdm+hal2(n0,seed,bytes) ... comb'd halton1R2b(n0)+random(seed,bytes)\n");
PE"                                        according to P. L'Ecuyer\n");
PE"            combHal(.)(.)(.)(.) ... all N Halton numbers combined together\n");
PE"                    (P. L'Ecuyer); same parameters as halton, incl. (rand)\n");
PE"                                   (BS algorithm)\n");
/*
PE"            rdmPhal2(seed,bytes) ... random() patched with halton1R2b(n0)\n");
						It is a disaster!!
*/
} /* --- end of:  describe_rand_gen() */


void out_RnGn_defaults(Nfict) int Nfict;
/*====================================*/
{
PE"Defaults for short specifications such as 'ranmar' or 'erand48(12', etc:\n");
PE"      r1\t\t\t\t\t   r2\n");
PE"   %s\t\t\t\t%s\n", DEFrandom[0], DEFrandom[1]);
PE"   %s\t\t\t\t\t%s\n", DEFrand[0], DEFrand[1]);
PE"   %s\t\t\t\t%s\n", DEFerand[0], DEFerand[1]);
PE"   %s\t\t\t\t%s\n", DEFranmar[0], DEFranmar[1]);
if(Nfict != 1)
	PE"   %s", DEFhalton[0]);
	else
	PE"   %s\t\t\t", DEFhalton_class);
PE"\t%s\n", DEFhalton[1]);
PE"   %s\t\t\t\t%s\n", DEFhalton1d[0], DEFhalton1d[1]);
PE"   %s\t\t\t\t%s\n", DEFhalton1i[0], DEFhalton1i[1]);
PE"   %s\t\t\t\t%s\n", DEFhalton1R[0], DEFhalton1R[1]);
PE"   %s\t\t\t\t%s\n", DEFhalton1R2b[0], DEFhalton1R2b[1]);
PE"   %s\t\t\t\t%s\n", DEFrdmHal2[0], DEFrdmHal2[1]);
PE"   %s\t\t\t\t\t%s\n\n", DEFcombHal[0], DEFcombHal[1]);
/* PE"   %s\t\t\t\t%s\n", DEFrdmPhal2[0], DEFrdmPhal2[1]);  */
} /* --- end of:  out_RnGn_defaults() */


/* --------------------------  write_RnGn():  ---------------------------- */

#define WRITE_RNGN(rg,State) switch( rg )\
{\
  case RANDOM:  WU(& State.random.seed, 1)\
		WI(& State.random.bytes, 1)\
		WI(& State.random.period, 1)\
		WI(State.random.state, State.random.bytes / sizeof(int))\
		break;\
  case RAND:    WU(& State.rand.seed, 1)\
		WI(& State.rand.last, 1)\
		break;\
  case ERAND48: WUS(State.erand48.ss0, 3)\
		WUS(State.erand48.ss, 3)\
		break;\
  case RANMAR:  WL(& State.ranmar.ij, 1)\
		WL(& State.ranmar.kl, 1)\
		WF(& State.ranmar.c, 1)\
		WF(& State.ranmar.cd, 1)\
		WF(& State.ranmar.cm, 1)\
		WF(State.ranmar.u, 97)\
		WL(& State.ranmar.i97, 1)\
		WL(& State.ranmar.j97, 1)\
		break;\
  case COMBHAL:\
  case HALTON:  WU(& State.halton.nBase, 1)\
		WU(& State.halton.iBase, 1)\
		WU(State.halton.Base, State.halton.nBase)\
		WU(State.halton.nra0, State.halton.nBase)\
		WU(State.halton.nrb0, State.halton.nBase)\
		WD(State.halton.nra1, State.halton.nBase)\
		WD(State.halton.nrb1, State.halton.nBase)\
		break;\
  case HALTON1d: WU(& State.halton1d.Base, 1)\
		 WU(& State.halton1d.nra0, 1)\
		 WU(& State.halton1d.nrb0, 1)\
		 WD(& State.halton1d.nra1, 1)\
		 WD(& State.halton1d.nrb1, 1)\
		 break;\
  case HALTON1i: WU(& State.halton1i.Base, 1)\
		 WU(& State.halton1i.nra0, 1)\
		 WU(& State.halton1i.nrb0, 1)\
		 WU(& State.halton1i.nra1, 1)\
		 WU(& State.halton1i.nrb1, 1)\
		 break;\
  case HALTON1R: WU(& State.halton1R.Base, 1)\
		 WU(& State.halton1R.n0, 1)\
		 WU(State.halton1R.dig, MAX_D_HALTON_R)\
		 WD(State.halton1R.r, MAX_D_HALTON_R)\
		 break;\
  case HALTON1R2b: WU(& State.halton1R2b.invN, 1)\
		   WU(& State.halton1R2b.n0, 1)\
		   break;\
  case RDMPHAL2:\
  case RDMHAL2: WU(& State.rdmHal2.seed, 1)\
		WI(& State.rdmHal2.bytes, 1)\
		WI(& State.rdmHal2.period, 1)\
		WI(State.rdmHal2.state, State.rdmHal2.bytes / sizeof(int))\
		WU(& State.rdmHal2.invN, 1)\
		WU(& State.rdmHal2.n0, 1)\
		break;\
}
/* see 1D.h for definitions of WI(), etc. */

void write_RnGn()
/*============*/
{
 WRITE_RNGN(rg1Flag, rg1_state)
 if(rg2Flag != SINGLE_STREAM) WRITE_RNGN(rg2Flag, rg2_state)
}
/* ----------------------- end of:   write_RnGn()  ------------------------ */


/* --------------------------  read_RnGn():  ---------------------------- */

#define READ_RNGN(rg,State) switch(rg)\
{\
  case RANDOM:  RU(& State.random.seed, 1)\
		RI(& State.random.bytes, 1)\
		RI(& State.random.period, 1)\
		RI(State.random.state, State.random.bytes / sizeof(int))\
		break;\
  case RAND:    RU(& State.rand.seed, 1)\
		RI(& State.rand.last, 1)\
		break;\
  case ERAND48: RUS(State.erand48.ss0, 3)\
		RUS(State.erand48.ss, 3)\
		break;\
  case RANMAR:  RL(& State.ranmar.ij, 1)\
		RL(& State.ranmar.kl, 1)\
		RF(& State.ranmar.c, 1)\
		RF(& State.ranmar.cd, 1)\
		RF(& State.ranmar.cm, 1)\
		RF(State.ranmar.u, 97)\
		RL(& State.ranmar.i97, 1)\
		RL(& State.ranmar.j97, 1)\
		break;\
  case COMBHAL:\
  case HALTON:  RU(& State.halton.nBase, 1)\
		RU(& State.halton.iBase, 1)\
		RU(State.halton.Base, State.halton.nBase)\
		RU(State.halton.nra0, State.halton.nBase)\
		RU(State.halton.nrb0, State.halton.nBase)\
		RD(State.halton.nra1, State.halton.nBase)\
		RD(State.halton.nrb1, State.halton.nBase)\
		break;\
  case HALTON1d: RU(& State.halton1d.Base, 1)\
		 RU(& State.halton1d.nra0, 1)\
		 RU(& State.halton1d.nrb0, 1)\
		 RD(& State.halton1d.nra1, 1)\
		 RD(& State.halton1d.nrb1, 1)\
		 break;\
  case HALTON1i: RU(& State.halton1i.Base, 1)\
		 RU(& State.halton1i.nra0, 1)\
		 RU(& State.halton1i.nrb0, 1)\
		 RU(& State.halton1i.nra1, 1)\
		 RU(& State.halton1i.nrb1, 1)\
		 break;\
  case HALTON1R: RU(& State.halton1R.Base, 1)\
		 RU(& State.halton1R.n0, 1)\
		 RU(State.halton1R.dig, MAX_D_HALTON_R)\
		 RD(State.halton1R.r, MAX_D_HALTON_R)\
		 break;\
  case HALTON1R2b: RU(& State.halton1R2b.invN, 1)\
		   RU(& State.halton1R2b.n0, 1)\
		   break;\
  case RDMPHAL2:\
  case RDMHAL2: RU(& State.rdmHal2.seed, 1)\
		RI(& State.rdmHal2.bytes, 1)\
		RI(& State.rdmHal2.period, 1)\
		RI(State.rdmHal2.state, State.rdmHal2.bytes / sizeof(int))\
		RU(& State.rdmHal2.invN, 1)\
		RU(& State.rdmHal2.n0, 1)\
		break;\
}
/* see 1D.h for definitions of RI(), etc. */

void read_RnGn(p) char *p;
/*======================*/
{
 READ_RNGN(rg1Flag, rg1_state)
 if(rg2Flag != SINGLE_STREAM) READ_RNGN(rg2Flag, rg2_state)
}
/* ----------------------- end of:   read_RnGn()  ------------------------ */


/* ---------------------------  out_RnGn_state():  ----------------------- */

#define PUT_HALTON(State,x)	PR"("); for(i = 0; i < j;  )\
                        { PR"%u", State.halton.x[i]);\
                          if(++i != j) PR","); else PR")");\
                        }\
			
#define PUT1_HALTON(State,Xe,x)	PR Xe );  for(i = 0; i < j; i++)\
	{ if(i > 0) if(i % 10 == 0) PR",\n           "); else PR",");\
	  PR"%5g", State.halton.x[i]);\
	}     PR"\n");\

#define OUT_RNGN(rg,State,Lbl) switch(rg)\
{\
  case RANDOM:  PR"random(%u,%d)\n", State.random.seed, State.random.bytes);\
		if(mode)\
		 { j = min(RANDOM_LEN, State.random.bytes / sizeof(int)) + 3;\
		   PR"    Current state:  ");\
		   for(i = 3, by = State.random.state; i < j ; i++, by++)\
			{ PR"%11d ", *by); if(!(i % 6)) PR"\n"); }\
		   PR"\n");\
		 }\
		break;\
  case RAND:    PR"rand(%u)\n", State.rand.seed);\
		if(mode) PR"    Current state: %d\n", State.rand.last);\
		break;\
  case ERAND48: PR"erand48(%hu,%hu,%hu)\n", State.erand48.ss0[0],\
			State.erand48.ss0[1], State.erand48.ss0[2]);\
		if(mode)\
		 PR"    Current state: %hu,%hu,%hu\n", State.erand48.ss[0],\
		   State.erand48.ss[1], State.erand48.ss[2]);\
		break;\
  case RANMAR:  PR"ranmar(%ld,%ld)\n", State.ranmar.ij, State.ranmar.kl);\
		if(mode)\
		 { PR"    Current state: i97 = %ld, j97 = %ld\n",\
		     State.ranmar.i97, State.ranmar.j97);\
	    PR"                   c = %.10g, cd = %.10g, cm = %.10g\n    ",\
		     State.ranmar.c, State.ranmar.cd, State.ranmar.cm);\
		   for(i = 0; i < 97;   )\
			{ PR"%.10g ", State.ranmar.u[i]);\
			  if(++i % 5 == 0) PR"\n    "); }\
		   PR"\n");\
		 }   break;\
  case COMBHAL: PR"combHal(%u)", State.halton.nBase); goto Lbl;\
  case HALTON:  PR"halton(%u)", State.halton.nBase);\
	     Lbl: j = min(State.halton.nBase, HALTON_MAXBASE);\
		PUT_HALTON(State,Base) PUT_HALTON(State,nra0)\
		PUT_HALTON(State,nrb0)  PR"\n");\
		if(mode)\
		 { PR"    Current state: iBase = %u\n", State.halton.iBase);\
		   PUT1_HALTON(State,"     nra1 = ",nra1)\
		   PUT1_HALTON(State,"     nrb1 = ",nrb1)\
		 }\
		break;\
  case HALTON1d: PR"halton1d(%u,%u,%u)\n", State.halton1d.Base,\
		   State.halton1d.nra0, State.halton1d.nrb0);\
		 if(mode) PR"    Current state: nra1 = %.20g, nrb1 = %.20g\n",\
			  State.halton1d.nra1, State.halton1d.nrb1);\
		 break;\
  case HALTON1i: PR"halton1i(%u,%u,%u)\n", State.halton1i.Base,\
		   State.halton1i.nra0, State.halton1i.nrb0);\
		 if(mode) PR"    Current state: nra1 = %u, nrb1 = %u\n",\
			  State.halton1i.nra1, State.halton1i.nrb1);\
		 break;\
  case HALTON1R: PR"halton1R(%u,%u)\n",State.halton1R.Base,State.halton1R.n0);\
		 if(mode) { int count0 = 0;\
			    for(i = MAX_D_HALTON_R - 1; i >= 0; i--)\
				if(dig[i]) break; else count0++;\
			    PR"    Current state: dig = ");\
			    for(i = 0; i < MAX_D_HALTON_R - count0; i++)\
				PR"%u ", State.halton1R.dig[i]);\
			    PR" %d*0\n    r = ", count0);\
			    count0 = 0;\
			    for(i = MAX_D_HALTON_R - 1; i >= 0; i--)\
				if(r[i]) break; else  count0++;\
			    for(i = 0; i < MAX_D_HALTON_R - count0; )\
				{ PR"%.10g ", State.halton1R.r[i]);\
				  if(++i % 5 == 0) PR"\n        "); }\
			    PR" %d*0\n", count0);\
			  }\
		 break;\
  case RDMHAL2: PR"rdm+hal2(%u,%u,%d)\n", State.rdmHal2.n0, State.rdmHal2.seed,\
					State.rdmHal2.bytes);\
		if(mode)\
		 { j = min(RANDOM_LEN, State.rdmHal2.bytes / sizeof(int)) + 1;\
		   PR"    Current state: invN = %u\n", State.rdmHal2.invN);\
		   for(i = 1, by = State.rdmHal2.state; i < j ; i++, by++)\
			{ PR"%11d ", *by); if(!(i % 6)) PR"\n"); }\
		   PR"\n");\
		 }\
		break;\
  case RDMPHAL2: PR"rdmPhal2(%u,%d)\n", State.rdmHal2.seed,\
					State.rdmHal2.bytes);\
		if(mode)\
		 { j = min(RANDOM_LEN, State.rdmHal2.bytes / sizeof(int)) + 1;\
		   PR"    Current state: invN = %u\n", State.rdmHal2.invN);\
		   for(i = 1, by = State.rdmHal2.state; i < j ; i++, by++)\
			{ PR"%11d ", *by); if(!(i % 6)) PR"\n"); }\
		   PR"\n");\
		 }\
		break;\
  case HALTON1R2b: PR"halton1R2b(%u)\n", State.halton1R2b.n0);\
		   if(mode)\
		   PR"    Current state: invN = %u\n", State.halton1R2b.invN);\
		   break;\
  case SINGLE_STREAM: PR"%s\n", Single_stream_string);\
}


void out_RnGn_state(mode,Listing) int mode; FILE *Listing;
/*======================================================*/
{
 /*  Current state of the selected RNG is printed only if mode != 0  */

 int i, j;
 int *by;

 if(mode) PR"\n");
 PR"1st RN generator: ");
 OUT_RNGN(rg1Flag, rg1_state, L1)
 PR"2nd RN generator: ");
 OUT_RNGN(rg2Flag, rg2_state, L2)
}
/* -------------------  end of:  out_RnGn_state()  ----------------------- */
