/*************************************************************************/
/********************** (C) 1992 by Miroslav Kolar ***********************

   Generator of the quasi-random Halton numbers.
   See: M. Berblinger & C. Schlier, Computer Phys. Commun. 66 (1991), 157

   nBase  bases  is used, subsequent calls to  halton()  use the bases
   periodically. That is, nBase calls to  halton()  generate one
   Halton vector.
**************************************************************************/

#include <stdio.h>

#define PR	(void)fprintf(Listing,
#define MAXBASE 10

extern FILE *Listing;

int nBase = 1, Base[MAXBASE] = {3},  iBase = 0, init = 0;
double nra1[MAXBASE] = {0.}, nrb1[MAXBASE] = {0.};
     /* nra1[] and nrb1[] always contain integers (see the above paper).
	Here we use double precision to increase the maximum length
	of the generated sequence beyond 2^31/max(Base[])
     */	

/* To initialize halton(), do at most MAXBASE calls to init_halton()
   If no call to init_halton() is made, the above initialization with a
   single basis value equal to 3 will be used, to prevent crash due to, as
   no check for initialization is done in halton() so as not to slow it down.

   The bases should be mutually prime, as small as possible.
*/

void reset_halton()
/*===============*/
{
 init = 0;
}

void init_halton(base) int base;
/*============================*/
{
 if( ! init) { nBase = 0; init = 1; }
 if(nBase >= MAXBASE) {
    PR"\nTo add another element to the set of Halton bases, MAXBASE must be\n");
    PR"increased in halton()\n\n");
    exit(1);
 }
 Base[nBase] = base;
 nra1[nBase] = 0.;
 nrb1[nBase] = 1.;
 ++nBase;
}


double halton()
/*===========*/
{
 register int base;
 register double nxb1;
 register double nxa1;

 base = Base[iBase];
 nxa1 = nrb1[iBase] - nra1[iBase];
 if(nxa1 == (double)1.) {
    nra1[iBase] = 1;
    nrb1[iBase] *= base;
 }
 else
 {  nxb1 = nrb1[iBase] / base;
    while(nxa1 <= nxb1) nxb1 /= base;
    nra1[iBase] = (1 + base) * nxb1 - nxa1;
 }
     {  double i = nrb1[iBase], j = 0;
	while(i > 1.) { i /= Base[iBase]; j++; }
	if(i != 1.) PR"%g  %g    %g  %d\n", nra1[iBase], nrb1[iBase], i, j);
     }
 nxb1 = nra1[iBase] / nrb1[iBase];
 ++iBase;  iBase %= nBase;
 return nxb1;
}
