/************************************************************************/
/*  This program tries to establish the period of various generators	*/
/*			defined in  RanGen.c				*/
/*									*/
/*			   M. Kolar 1992				*/
/************************************************************************/

#include <errno.h>
#include <sys/times.h>
#include "1D.h"

void  err_fread(p) char *p; {}  /* needed for RanGen.c but not used here */
FILE *Listing = stdout, *Output = stdout, *Density = stdout, *Input;

/*-----------------------------------------------------------------------*/
/* Currently available random number generators are defined in RanGen.c. */
/* They require the following declarations:				 */

int rg1Flag, rg2Flag;  /* indicate the two selected generators for
			 particle movement and configuration acceptance	*/
char *rg1 = "\0", *rg2 = SINGLE_STREAM_STRING;
	 /* pointers to verbal descriptions of the two RNG's */

double randNum1[1], randNum2[1];
	/* buffers to store N_fict random numbers generated by the 
	   two generators; these numbers are just enough for one MC cycle */

extern void out_RnGn_state();  /* formated output of the state vectors  */

extern int init_rand_gen();  /* selects and initializes the two generators */

extern double (*RnGn1)(), (*RnGn2)();
	  /* pointers to the two selected RNG */

/*	End of definitions for the file RanGen.c			 */
/*-----------------------------------------------------------------------*/



main(ac, av) int ac; char *av[];
/*============================*/
{
register int j;
register int i;
register double rNum;
int k, l, o_last = -10, o_count = 0;

clock_t times();
double usLast, sysLast, uTo, sTo;
struct tms cpuT;

#define BATCH	5
#define FIRST	2000
#define LAST    (FIRST + BATCH - 1)
#define BATCH_1 (BATCH - 1)
#define MAXN	200000000	/* should be larger than FIRST */ 
#define GNUM	6	/* Number of generators from RG[] to be tested */

double batchRN[BATCH];
int    ind[BATCH];
static char *RG[] = { "ranmar(1874,8371)",     "ranmar(327,18534)",
		      "random(4123,256)",      "random(1,256)",
		      "halton(1)",	       "halton(1)(7)",
		      "erand48(1745,719,3)",   "rand(1)",
		      "erand48(75,813,47345)", "rand(43123)",
		      "halton(5)",	       "halton(10)"		};


PR"\nStarted:  "); PR DateAndTime());
PR"\n\n            Period of random number generators\n");
PR"            ==================================\n");
PR"               Written by M. Kolar (C) 1992\n\n\n");


for(l = 0; l < BATCH; l++) ind[l] = -1;
(void)times(&cpuT);
usLast = (double)cpuT.tms_utime / 60;
sysLast = (double)cpuT.tms_stime / 60;
PR"CPU time from start: user  %g sec, system  %g sec\n\n", usLast, sysLast);

for(k = 0; k < GNUM; k++)
 {
   rg1 = RG[k];
   PR"==================>  Testing  %s  <==================\n", rg1);
   if( init_rand_gen(&rg1, &rg2, 0, 1) ) { PR"\n\n"); continue; }
   out_RnGn_state(1,Listing);
   PR"\n");
   for(i = 0; i < FIRST; i++) (void)(*RnGn1)();
   for(i = FIRST; i <= LAST; i++) batchRN[i - FIRST] = (*RnGn1)();
   for(i = LAST + 1; i < MAXN; i++)
	{  rNum = (*RnGn1)();
	   if(rNum < 0. || rNum >= 1.)
		{ PR"!!!!!! Out of bounds: %d %.20g\n", i, rNum);
		  if(o_last == i - 1) o_count++;
		  if(o_count >= 500) goto Break;
		}
	   if(rNum == 0.) PR"%d %.20g\n", i, rNum);
	   for(j = 0; j < BATCH; j++)
		{ if(rNum == batchRN[j])
		   { PR"!!! %dth random number is equal to the %dth number\n",
				i, j + FIRST);
		     for(l = 0; l < BATCH_1; l++) ind[l] = ind[l+1];
		     ind[BATCH_1] = i;
		     if(j == BATCH_1)
			  { for(l = 1; l < BATCH; l++)
				if(ind[BATCH_1 - l] != i - l) break;
			    if(l == BATCH) goto Break;
			  }
		     break;
		   }	
		}
	}
 Break:
   if(i == MAXN) PR"\nPeriod of this generator is larger or equal to %d\n",
			MAXN);
   PR"\n");
   out_RnGn_state(1,Listing);
   PR"\nReal time:  "); PR DateAndTime());
   (void)times(&cpuT);
   uTo = (double)cpuT.tms_utime / 60;
   sTo = (double)cpuT.tms_stime / 60;
   PR"\nCPU: user  %g sec, system  %g sec, total  %g sec\n",
	uTo - usLast , sTo - sysLast, uTo + sTo - (usLast + sysLast));
   usLast = uTo;
   sysLast = sTo;
   PR"\n\n\n");
   (void)fflush(Listing);
 }

PR"Finished:  "); PR DateAndTime());
PR"\n\nWhole job: user  %g sec, system  %g sec, total  %g sec\n",
					uTo, sTo, uTo + sTo);
PR"\n");

}
